# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this 
# file except in compliance with the License. You may obtain a copy of the License at
# 
# https://opensource.org/licenses/BSD-3-Clause
# 
# Unless required by applicable law or agreed to in writing, software distributed 
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''QPath中间层
'''


from tuia.qpathparser import QPathParser
QPathParser.INT_TYPE_PROPNAMES = ['MAXDEPTH']

class QPath(object):
    '''QPath
    '''
    def __init__(self, qpath_str):
        self._qpath_str = qpath_str
    
    def __str__(self):
        return self._qpath_str
    
    @property
    def parsed_qpath(self):
        '''解析后的内容
        '''
        result = QPathParser().parse(self._qpath_str)[0]
        for it in result:
            for key in it:
                if not key in ['Id', 'Text', 'Type', 'Visible', 'Desc', 'MaxDepth', 'Instance']:
                    if not key.startswith('Field_') and not key.startswith('Method_'):
                        raise RuntimeError('QPath(%s) keyword error: %s' % (self._qpath_str, key))
        return result
    
    @property
    def _parsed_qpath(self):
        return self.parsed_qpath

if __name__ == '__main__':
    pass



# -*- coding: UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''Android测试桩代理
'''
from __future__ import print_function

import json
import re
import os
import six
import sys
import tempfile
import threading
import time

from qt4a.androiddriver.clientsocket import DirectAndroidSpyClient
from qt4a.androiddriver.devicedriver import DeviceDriver
from qt4a.androiddriver.util import logger, general_encode, SocketError, AndroidSpyError, ControlAmbiguousError, ControlExpiredError, ProcessExitError, QT4ADriverNotInstalled, Mutex, extract_from_zipfile


class EnumCommand(object):
    '''所有支持的命令字
    '''
    CmdHello = "Hello"
    CmdEnableDebug = "EnableDebug"
    CmdGetControl = "GetControl"
    CmdGetParent = "GetParent"
    CmdGetChildren = "GetChildren"
    CmdGetControlText = "GetControlText"
    CmdSetControlText = "SetControlText"
    CmdGetTextRect = "GetTextBoundingRect"
    CmdGetControlRect = "GetControlRect"
    CmdGetControlType = "GetControlType"
    CmdGetControlVisibility = "GetControlVisibility"
    CmdGetControlChecked = "GetControlChecked"
    CmdGetControlProgress = "GetControlProgress"
    CmdGetControlGravity = "GetControlGravity"
    CmdCaptureControl = "CaptureControl"
    CmdSendKey = "SendKey"
    CmdClick = "Click"
    CmdDrag = "Drag"
    CmdEnableSoftInput = "EnableSoftInput"
    CmdCloseActivity = "CloseActivity"
    CmdEvalScript = "EvalScript"
    CmdGetCurrentView = "GetCurrentView"
    CmdGetControlScrollRect = "GetControlScrollRect"
    CmdGetListViewInfo = "GetListViewInfo"
    CmdGetControlBackground = "GetControlBackground"
    CmdGetControlImageResource = "GetControlImageResource"
    CmdGetSelectedTabIndex = "GetSelectedTabIndex"
    CmdGetStaticFieldValue = "GetStaticFieldValue"
    CmdGetObjectFieldValue = "GetObjectFieldValue"
    CmdCallStaticMethod = "CallStaticMethod"
    CmdCallObjectMethod = "CallObjectMethod"
    CmdCallExternalMethod = "CallExternalMethod"  # 调用外部jar包中的方法
    CmdSetActivityPopup = "SetActivityPopup"
    CmdSetThreadPriority = "SetThreadPriority"
    CmdSetWebViewDebuggingEnabled = "SetWebViewDebuggingEnabled"


def install_qt4a_helper(adb, root_path):
    from qt4a.androiddriver.util import AndroidPackage, version_cmp
    qt4a_helper_package = 'com.test.androidspy'
    apk_path = os.path.join(root_path, 'QT4AHelper.apk')
    if adb.get_package_path(qt4a_helper_package):
        # 判断版本
        installed_version = adb.get_package_version(qt4a_helper_package)
        package = AndroidPackage(apk_path)
        install_version = package.version
        if version_cmp(install_version, installed_version) > 0:
            adb.install_apk(apk_path, True)
    else:
        adb.install_apk(apk_path)


def copy_android_driver(device_id_or_adb, force=False, root_path=None, enable_acc=True):
    '''测试前的测试桩拷贝
    '''
    from qt4a.androiddriver.adb import ADB

    if isinstance(device_id_or_adb, ADB):
        adb = device_id_or_adb
    else:
        adb = ADB.open_device(device_id_or_adb)

    if not root_path:
        current_path = os.path.abspath(__file__)
        if not os.path.exists(current_path) and '.egg' in current_path:
            # in egg
            egg_path = current_path
            while not os.path.exists(egg_path):
                egg_path = os.path.dirname(egg_path)
            assert(egg_path.endswith('.egg'))
            root_path = os.path.join(tempfile.mkdtemp(), 'tools')
            extract_from_zipfile(
                egg_path, 'qt4a/androiddriver/tools', root_path)
        else:
            root_path = os.path.join(os.path.dirname(
                os.path.abspath(__file__)), 'tools')
    dst_path = '/data/local/tmp/qt4a/'

    current_version_file = os.path.join(root_path, 'version.txt')
    f = open(current_version_file, 'r')
    current_version = int(f.read())
    f.close()

    if not force:
        version_file = dst_path + 'version.txt'
        version = adb.run_shell_cmd('cat %s' % version_file)

        if version and not 'No such file or directory' in version and current_version <= int(version):
            install_qt4a_helper(adb, root_path) # 避免QT4A助手被意外删除的情况
            # 不需要拷贝测试桩
            logger.warn('忽略本次测试桩拷贝：当前版本为%s，设备中版本为%s' %
                        (current_version, int(version)))
            return

    try:
        adb.chmod(dst_path[:-1], '777')
    except:
        pass

    rooted = adb.is_rooted()

    cpu_abi = adb.get_cpu_abi()
    print('Current CPU arch: %s' % cpu_abi)
    use_pie = False
    if adb.get_sdk_version() >= 21 and cpu_abi != 'arm64-v8a':
        use_pie = True

    file_list = [os.path.join(cpu_abi, 'droid_inject'), os.path.join(cpu_abi, 'libdexloader.so'), os.path.join(
        cpu_abi, 'screenkit'), os.path.join(cpu_abi, 'libandroidhook.so'), 'inject', 'AndroidSpy.jar', 'SpyHelper.jar', 'SpyHelper.sh']

    if cpu_abi == 'arm64-v8a':
        file_list.append(os.path.join(cpu_abi, 'droid_inject64'))
        file_list.append(os.path.join(cpu_abi, 'libdexloader64.so'))
        file_list.append('inject64')
    if adb.get_sdk_version() >= 21:
        file_list.append(os.path.join(cpu_abi, 'libandroidhook_art.so'))

    if rooted and adb.is_selinux_opened():
        # 此时如果还是开启状态说明关闭selinux没有生效,主要是三星手机上面
        adb.run_shell_cmd('rm -r %s' % dst_path, True)
        # adb.run_shell_cmd('chcon u:object_r:shell_data_file:s0 %slibdexloader.so' % dst_path, True)  # 恢复文件context，否则拷贝失败
        # adb.run_shell_cmd('chcon u:object_r:shell_data_file:s0 %slibandroidhook.so' % dst_path, True)

    for file in file_list:
        file_path = os.path.join(root_path, file)
        if use_pie and not '.' in file and os.path.exists(file_path + '_pie'):
            file_path += '_pie'
        if not os.path.exists(file_path):
            continue
        save_name = os.path.split(file)[-1]
        if save_name.endswith('_art.so'):
            save_name = save_name.replace('_art', '')
        adb.push_file(file_path, dst_path + save_name)

    adb.chmod('%sdroid_inject' % dst_path, 755)
    adb.chmod('%sinject' % dst_path, 755)
    adb.chmod('%sscreenkit' % dst_path, 755)
    adb.run_shell_cmd('ln -s %sscreenkit %sscreenshot' % (dst_path, dst_path))

    if cpu_abi == 'arm64-v8a':
        adb.chmod('%sdroid_inject64' % dst_path, 755)
        adb.chmod('%sinject64' % dst_path, 755)

    try:
        print(adb.run_shell_cmd('rm -R %scache' %
                                dst_path, rooted))  # 删除目录 rm -rf
    except RuntimeError as e:
        logger.warn('%s' % e)
    # logger.info(adb.run_shell_cmd('mkdir %scache' % (dst_path), True)) #必须使用root权限，不然生成odex文件会失败

    adb.mkdir('%scache' % (dst_path), 777)

    install_qt4a_helper(adb, root_path)

    version_file_path = os.path.join(root_path, 'version.txt')
    dst_version_file_path = dst_path + os.path.split(version_file_path)[-1]
    adb.push_file(version_file_path, dst_version_file_path + '.tmp')  # 拷贝版本文件

    if rooted and adb.is_selinux_opened():
        # 此时如果还是开启状态说明关闭selinux没有生效,主要是三星手机上面
        # 获取sdcars context
        if adb.get_sdk_version() >= 23:
            sdcard_path = adb.get_sdcard_path()
            result = adb.run_shell_cmd('ls -Z %s' % sdcard_path)
            # u:object_r:media_rw_data_file:s0 u:object_r:rootfs:s0
            pattern = re.compile(r'\s+(u:object_r:.+:s0)\s+')
            ret = pattern.search(result)
            if not ret:
                raise RuntimeError('get sdcard context failed: %s' % result)
            context = ret.group(1)
            logger.info('sdcard context is %s' % context)
            adb.run_shell_cmd('chcon %s %s' %
                              (context, dst_path), True)  # make app access
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %sSpyHelper.jar' % dst_path, True)
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %sSpyHelper.sh' % dst_path, True)
            # 不修改文件context无法加载so
            adb.run_shell_cmd(
                'chcon u:object_r:system_file:s0 %slibdexloader.so' % dst_path, True)
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %slibandroidhook.so' % dst_path, True)
            adb.run_shell_cmd('chcon %s %sAndroidSpy.jar' %
                              (context, dst_path), True)
            adb.run_shell_cmd('chcon %s %scache' % (context, dst_path), True)
        else:
            # 不修改文件context无法加载so
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %slibdexloader.so' % dst_path, True)
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %slibandroidhook.so' % dst_path, True)
            adb.run_shell_cmd(
                'chcon u:object_r:app_data_file:s0 %scache' % dst_path, True)

    if rooted:
        if adb.get_sdk_version() < 24:
            # 7.0以上发现生成的dex与运行时生成的dex有差别，可能导致crash
            logger.info(adb.run_shell_cmd('sh %sSpyHelper.sh loadDex %sAndroidSpy.jar %scache' % (
                dst_path, dst_path, dst_path), rooted))
            adb.chmod('%scache/AndroidSpy.dex' % dst_path, 666)
    else:
        if not 'usage:' in adb.run_shell_cmd('sh %sSpyHelper.sh' % dst_path):
            adb.mkdir('%scache/dalvik-cache' % dst_path, 777)

    if rooted and adb.is_selinux_opened() and adb.get_sdk_version() >= 23:
        # 提升权限
        try:
            adb.list_dir('/system/bin/app_process32')
        except RuntimeError:
            adb.copy_file('/system/bin/app_process',
                          '%sapp_process' % dst_path)
        else:
            adb.copy_file('/system/bin/app_process32',
                          '%sapp_process' % dst_path)
        adb.chmod('%sapp_process' % dst_path, 755)
        adb.run_shell_cmd(
            'chcon u:object_r:system_file:s0 %sapp_process' % dst_path, True)

    adb.run_shell_cmd('mv %s %s' % (dst_version_file_path +
                                    '.tmp', dst_version_file_path), rooted)

    # 同步手机时间
    device_driver = DeviceDriver(adb)
    try:
        input_method = 'com.test.androidspy/.service.QT4AKeyboardService'
        device_driver.modify_system_setting(
            'secure', 'enabled_input_methods', input_method)
        device_driver.modify_system_setting(
            'secure', 'default_input_method', input_method)
        if enable_acc:
            device_driver.modify_system_setting('secure', 'enabled_accessibility_services',
                                                'com.test.androidspy/com.test.androidspy.service.QT4AAccessibilityService')
            device_driver.modify_system_setting(
                'secure', 'accessibility_enabled', 1)
    except:
        logger.exception('set default input method failed')
    try:
        device_driver.modify_system_setting('system', 'time_12_24', 24)
        device_driver.modify_system_setting(
            'system', 'screen_off_timeout', 600 * 1000)
    except:
        logger.exception('set system time failed')


class AndroidDriver(object):
    '''
    '''

    qt4a_path = '/data/local/tmp/qt4a'

    def __init__(self, device_driver, process_name, addr='127.0.0.1'):
        self._device_driver = device_driver
        self._adb = device_driver.adb
        self._process_name = process_name
        self._addr = addr
        self._process = {}  # 保存原始PID用于判断是否是新进程
        self._sync_lock = {}  # 同步锁
        self._max_block_time = 30  # 最长锁时间,防止卡死
        self._lock = threading.Lock()  # 创建AndroidDriver实例的互斥锁
        self._is_init = False
        self._client = None

    @staticmethod
    def create(process_name, device_or_driver):
        '''创建Driver实例

        :param process_name: 目标进程名
        :type  process_name: string
        :param device:       设备实例
        :type  device:       Device or DeviceDriver
        '''
        if hasattr(device_or_driver, '_device_driver'):
            device_driver = device_or_driver._device_driver
        else:
            device_driver = device_or_driver
        host_addr = '127.0.0.1'
        if not device_driver._is_local_device:
            host_addr = device_driver.adb.host_name
        return AndroidDriver(device_driver, process_name, addr=host_addr)

    @staticmethod
    def get_process_name_hash(process_name, device_id):
        '''根据进程名和设备id计算端口值
                结果范围：[5100, 5200)
        TODO: hash冲突的解决
        '''
        result = 0
        start_port = 15000
        port_range = 1000
        text = device_id + process_name
        for i in range(len(text)):
            c = text[i]
            asc = ord(c)
            result = 31 * result + asc

        if result > port_range:
            result %= port_range
        return result + start_port

    def _create_client(self):
        '''创建新的Client实例
        '''
        sock = self._adb.create_tunnel(self._process_name, 'localabstract')
        if sock == None:
            return None
        return DirectAndroidSpyClient(sock, timeout=360)

    def _safe_init_driver(self):
        '''多线程安全的初始化测试桩
        '''
        if not self._is_init:
            with Mutex(self._lock):
                if not self._is_init:
                    self._init_driver()
                    self._is_init = True

    def _init_driver(self):
        '''初始化测试桩
        '''
        self._client = self._create_client()
        if self._client != None:
            # 字段赋值
            self._process['name'] = self._process_name
            self._process['id'] = 0  # process id may change
            if self.hello() != None:
                self._process['id'] = self._adb.get_pid(self._process_name)
                return

        timeout = 20
        time0 = time.time()
        proc_exist = False
        while time.time() - time0 < timeout:
            if not proc_exist:
                pid = self._adb.get_pid(self._process_name)
                if pid > 0:
                    proc_exist = True
                    self._process['name'] = self._process_name
                    self._process['id'] = pid
                    break

            time.sleep(1)

        if not proc_exist:
            raise RuntimeError('进程：%s 在%d秒内没有出现' %
                               (self._process_name, timeout))

        inject_file = 'inject'
        if self._adb.is_app_process64(pid if self._adb.is_rooted() else self._process_name):
            # 64 bit process
            inject_file += '64'
        timeout = 30

        try:
            if self._adb.is_art():
                # Android 5.0上发现注入容易导致进程退出
                self._wait_for_cpu_low(20, 10)

            time0 = time.time()
            cmdline = '%s/%s %s' % (self._get_driver_root_path(),
                                    inject_file, self._process_name)
            while time.time() - time0 < timeout:
                if self._adb.is_rooted():
                    ret = self._adb.run_shell_cmd(
                        cmdline, True, timeout=120, retry_count=1)
                else:
                    ret = self._adb.run_as(
                        self._process_name, cmdline, timeout=120, retry_count=1)
                logger.debug('inject result: %s' % ret)
                if 'not found' in ret:
                    raise QT4ADriverNotInstalled(
                        'QT4A driver damaged, please reinstall QT4A driver')
                if 'Inject Success' in ret:
                    break
                elif 'Operation not permitted' in ret:
                    # 可能是进程处于Trace状态
                    pid = self._adb.get_pid(self._process_name)
                    status = self._adb.get_process_status(pid)
                    tracer_pid = int(status['TracerPid'])
                    if tracer_pid > 0:
                        if int(status['PPid']) == tracer_pid:
                            # 使用TRACEME方式防注入
                            raise Exception('应用使用了防注入逻辑,注入失败')
                        logger.warn('TracerPid is %d' % tracer_pid)
                        self._adb.kill_process(tracer_pid)
                time.sleep(1)

        except RuntimeError as e:
            logger.exception('%s\n%s' % (e, self._adb.run_shell_cmd('ps')))
            if self._adb.is_rooted():
                logger.info(self._adb.dump_stack(self._process_name))
            raise e
        timeout = 10
        time0 = time.time()
        while time.time() - time0 < timeout:
            if self._client == None:
                self._client = self._create_client()
            if self._client != None and self.hello() != None:
                return
            time.sleep(0.1)
        raise RuntimeError('连接测试桩超时')

    def _get_driver_root_path(self):
        '''获取驱动文件根目录
        '''
        if self._adb.is_rooted():
            return self.__class__.qt4a_path
        package_name = self._process_name
        if ':' in package_name:
            package_name = package_name.split(':')[0]
        if not self._device_driver.is_debug_package(package_name):
            raise NotImplementedError(
                'Non root device doesn\'t support release package')
        root_path = '/data/data/%s/qt4a' % package_name  # 使用应用目录作为根目录
        result = self._adb.run_as(package_name, 'ls -l %s/inject' % root_path)
        if not 'No such file or directory' in result:
            return root_path

        # 拷贝驱动文件
        self._adb.run_as(package_name, 'mkdir %s' % root_path)
        self._adb.run_as(package_name, 'mkdir %s/cache' % root_path)
        file_list = ['AndroidSpy.jar']  # , 'cache/AndroidSpy.dex'
        if self._adb.is_app_process64(package_name):
            file_list.extend(
                ['droid_inject64', 'inject64', 'libdexloader64.so'])
        else:
            file_list.extend(
                ['droid_inject', 'inject', 'libdexloader.so', 'libandroidhook.so'])
        for filename in file_list:
            self._adb.run_as(package_name, 'cp %s/%s %s/%s' %
                             (self.__class__.qt4a_path, filename, root_path, filename))
        # self._device.adb.run_as(package_name, 'sh %s/SpyHelper.sh loadDex %s/AndroidSpy.jar %s/cache' % (self.__class__.qt4a_path, self.__class__.qt4a_path, root_path))
        return root_path

    def _wait_for_cpu_low(self, max_p_cpu, max_t_cpu, timeout=20, interval=0.5):
        '''等待待注入进程CPU使用率降低到max_p_cpu，线程CPU使用率降低到max_t_cpu

        :param max_p_cpu: 进程占用的CPU
        :type max_p_cpu:  int
        :param max_t_cpu: 线程占用的CPU
        :type max_t_cpu:  int
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            ret = self._adb.get_process_cpu(self._process_name)
            if ret != None:
                p_cpu, t_cpu = ret
                logger.debug('current cpu: %d, %d' % (p_cpu, t_cpu))
                if p_cpu < max_p_cpu and t_cpu < max_t_cpu:
                    return
            time.sleep(interval)

    def close(self):
        '''关闭连接
        '''
        if self._client:
            self._client.close()
            self._client = None

    def _get_lock(self, thread_id):
        '''获取锁
        '''
        if not thread_id in self._sync_lock:
            self._sync_lock[thread_id] = threading.Event()
            self._sync_lock[thread_id].set()
        return self._sync_lock[thread_id]

    def suspend_thread(self, thread_id, max_wait_time=30):
        '''阻塞
        '''
        self._max_block_time = max_wait_time
        self._get_lock(thread_id).clear()  # Reset the internal flag to false

    def resume_thread(self, thread_id):
        '''解锁
        '''
        self._get_lock(thread_id).set()  # Set the internal flag to true

    def _wait_for_event(self, thread_id, timeout):
        '''等待事件
        '''
        if self._get_lock(thread_id).is_set():
            return
        time0 = time.time()
        ret = self._get_lock(thread_id).wait(timeout)
        if ret:
            cost_time = time.time() - time0
            logger.debug('thread %s wait %s S' % (thread_id, cost_time))
        else:
            # 超时
            logger.warn('thread %s wait for sync event timeout' % thread_id)
            self.resume_thread(thread_id)  # 防止出错后一直等待
        return ret

    def send_command(self, cmd_type, **kwds):
        '''发送命令
        '''
        curr_thread_id = threading.current_thread().ident
        self._wait_for_event(curr_thread_id, self._max_block_time)

        if cmd_type != EnumCommand.CmdHello:
            self._safe_init_driver()  # 确保测试桩连接正常

        result = self._client.send_command(cmd_type, **kwds)
        if result == None:
            pid = self._adb.get_pid(self._process['name'])
            if pid > 0 and pid != self._process['id']:
                # 新进程，需要重新注入
                logger.info('process %s id changed: %d %d' %
                            (self._process['name'], self._process['id'], pid))
                self._is_init = False  # 需要重新初始化
                self._process['id'] = pid
                return self.send_command(cmd_type, **kwds)
            elif pid == 0:
                raise ProcessExitError('被测进程已退出，确认是否发生Crash')
            elif cmd_type != EnumCommand.CmdHello:
                # hello包不重试
                logger.debug('socket error, try to reconnect')
                for _ in range(3):
                    # 为防止由于设备短暂失联导致的连接断开，这里调一次adb forward
                    self._client = self._create_client()
                    result = self._client.send_command(cmd_type, **kwds)
                    if result != None:
                        return result
                raise SocketError('Connect Failed')
            else:
                raise SocketError('Connect Failed')
        if 'Error' in result:
            if result['Error'] == u'控件已失效':
                raise ControlExpiredError(result['Error'])
            elif result['Error'] == u'控件类型错误':
                control_type = self.get_control_type(kwds['Control'])
                raise TypeError('%s,当前控件类型为:%s' %
                                (result['Error'], control_type))
            else:
                raise AndroidSpyError(result['Error'])
        return result

    def hello(self):
        '''确认Server身份
        '''
        for _ in range(3):
            try:
                result = self.send_command(EnumCommand.CmdHello)
            except SocketError:
                time.sleep(1)
                continue
            except AndroidSpyError:
                logger.exception('init error')
                time.sleep(2)
                continue
            # self._enable_debug()
            if not 'Result' in result:
                raise RuntimeError('Server error')
            items = result['Result'].split(':')
            if len(items) > 0 and items[-1].isdigit():
                if self._process['id'] and int(items[-1]) != self._process['id']:
                    raise RuntimeError(
                        'Server pid not match %s' % result['Result'])
            return result['Result']

    def _enable_debug(self):
        '''启用Debug模式
        '''
        self.send_command(EnumCommand.CmdEnableDebug)

    def get_control(self, activity, parent, locator, get_position=False):
        '''获取控件hashcode
        '''
        kwds = {}
        if activity:
            kwds['Activity'] = activity
        if parent:
            kwds['Parent'] = parent
        kwds['Locator'] = locator
        if get_position:
            kwds['GetPosition'] = get_position
        try:
            result = self.send_command(EnumCommand.CmdGetControl, **kwds)
        except ControlExpiredError as e:
            raise e
        except AndroidSpyError as e:
            err_msg = e.args[0].strip()
            err_msg = general_encode(err_msg)
            if '找到重复控件' in err_msg or 'Multiple controls found' in err_msg:
                qpath = '/'.join(' && '.join([(key + it[key][0] + '"' + str(
                    it[key][1]) + '"') for key in it.keys()]) for it in locator)
                qpath = general_encode(qpath)
                err_msg = '定位到重复控件：/%s\n%s' % (qpath, '\n'.join(err_msg.split('\n')[1:]))
                raise ControlAmbiguousError(err_msg)
            elif '查找控件失败' in err_msg or 'Control not found' in err_msg:
                if get_position and ':' in err_msg:
                    _, pos = err_msg.split(':')
                    return int(pos)
            return 0
        if 'Result' in result:
            return result['Result']
        return 0

    def get_parent(self, control):
        '''获取父控件
        '''
        result = self.send_command(EnumCommand.CmdGetParent, Control=control)
        return result['Result']

    def get_children(self, control):
        '''获取子节点列表
        '''
        result = self.send_command(EnumCommand.CmdGetChildren, Control=control)
        return result['Result']

    def get_control_text(self, control, html_format=False):
        '''获取控件文本
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlText, Control=control, UseHtml=html_format)
        result = result['Result']
        if html_format:
            # 处理html编码
            from six.moves.html_parser import HTMLParser
            result = HTMLParser().unescape(result)
        return result

    def set_control_text(self, control, text):
        '''设置控件文本
        '''
        result = self.send_command(
            EnumCommand.CmdSetControlText, Control=control, Text=text)
        return result['Result']

    def get_text_rect(self, control, start_offset, end_offset):
        '''获取文本框中指定字符串的区域坐标
        '''
        result = self.send_command(
            EnumCommand.CmdGetTextRect, Control=control, StartOffset=start_offset, EndOffset=end_offset)
        return result['Result']

    def get_control_type(self, control, include_baseclass=False):
        '''获取控件类型
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlType, Control=control, IncludeBaseClass=include_baseclass)
        if not include_baseclass:
            return result['Result']
        else:
            return result['Result'].split(',')

    def get_control_rect(self, control):
        '''获取控件位置信息
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlRect, Control=control)
        return result['Result']

    def get_control_visibility(self, control):
        '''获取控件可见性
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlVisibility, Control=control)
        return result['Result']

    def get_control_checked(self, control):
        '''获取CommpandButton控件是否选择
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlChecked, Control=control)
        return result['Result']

    def get_control_gravity(self, control):
        '''获取控件位置信息
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlGravity, Control=control)
        return result['Result']

    def get_control_progress(self, control):
        '''获取ProgressBar控件的进度
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlProgress, Control=control)
        return result['Result']

    def get_control_background(self, control):
        '''获取控件背景资源id
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlBackground, Control=control)
        return result['Result']

    def get_control_scroll_rect(self, control):
        '''获取控件滚动区域范围
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlScrollRect, Control=control)
        return result['Result']

    def get_control_image_resource(self, control):
        '''获取控件Image资源信息
        '''
        result = self.send_command(
            EnumCommand.CmdGetControlImageResource, Control=control)
        return result['Result']

    def get_listview_info(self, control, timeout=20):
        '''获取AbsListView类型的相关信息
        '''
        time0 = time.time()
        while time.time() - time0 < timeout:
            # 过滤掉一些明显错误的数据
            result = self.send_command(
                EnumCommand.CmdGetListViewInfo, Control=control)
            result = result['Result']
            if (result['Count'] > 0 and result['FirstPosition'] >= result['Count']) or \
                    result['LastPosition'] > result['Count']:  # 最后一项>=总项数 {"LastPosition":0,"FirstPosition":0,"Count":0}
                time.sleep(0.03)
                continue
            return result
        raise RuntimeError('获取ListView控件信息错误')

    def get_selected_tab_index(self, control):
        '''获取TabWidget控件当前所选项
        '''
        result = self.send_command(
            EnumCommand.CmdGetSelectedTabIndex, Control=control)
        return result['Result']

    def capture_control(self, control):
        result = self.send_command(
            EnumCommand.CmdCaptureControl, Control=control)
        data = result['Result']
        # print (data)
        import base64
        data = base64.decodestring(data)
        return data

    def send_key(self, key_list):
        '''发送按键，只允许单个按键或组合键
        '''
        from qt4a.androiddriver.util import KeyCode
        if isinstance(key_list, str):
            return self.send_keys(key_list)
        elif isinstance(key_list, int):
            # 将整型按键转为list
            key_list = [key_list]
        if len(key_list) == 0 or len(key_list) > 3:
            raise RuntimeError('按键错误')
        keys = ''
        for key in key_list:
            keys += str(key) + ','
        keys = keys[:-1]
        if keys == str(KeyCode.KEYCODE_HOME) or keys == str(KeyCode.KEYCODE_BACK):  # Home键需要特殊处理
            return self._device_driver.send_key(int(keys))
        else:
            return self.send_command(EnumCommand.CmdSendKey, Keys=keys)

    def send_keys(self, text):
        '''发送字符串，目前不支持中文
        '''
        from qt4a.androiddriver.util import KeyCode
        key_list = KeyCode.get_key_list(text)
        for key in key_list:
            self.send_key(key)

    def click(self, control, x, y, sleep_time=0):
        if x < 0 or y < 0:
            raise RuntimeError('坐标错误：(%d, %d)' % (x, y))

        try:
            if not control:
                result = self.send_command(
                    EnumCommand.CmdClick, X=x, Y=y, SleepTime=int(sleep_time * 1000))
            else:
                result = self.send_command(
                    EnumCommand.CmdClick, Control=control, X=x, Y=y, SleepTime=int(sleep_time * 1000))
            # print (result['Result'])
            return result['Result']
        except AndroidSpyError as e:
            if str(e).find('java.lang.SecurityException') >= 0:
                # 有时操作过快会出现该问题
                return False
            elif 'ErrControlNotShown' in str(e):
                return False
            elif str(e) == '控件已失效':
                # 认为点击成功了
                pic_name = '%s.png' % int(time.time())
                logger.warn('%s %s' % (e, pic_name))
                self._device_driver.take_screen_shot(pic_name)
                return True
            else:
                raise e

    def drag(self, x1, y1, x2, y2, count=5, wait_time=40, send_down_event=True, send_up_event=True):
        '''滑动
        :param x1: 起始横坐标
        :type x1:  int
        :param y1: 起始纵坐标
        :type y1:  int
        :param x2: 终点横坐标
        :type x2:  int
        :param y2: 终点纵坐标
        :type y2:  int
        :param count: 步数
        :type count: int
        :param wait_time: 滑动过程中每步之间的间隔时间，ms
        :type wait_time: int
        :param send_down_event: 是否发送按下消息
        :type send_down_event: boolean
        :param send_up_event: 是否发送弹起消息
        :type send_up_event: boolean 
        '''
        if y1 != y2 and abs(y1 - y2) < 60:  # 此时滑动距离会很短
            # 保证最短滑动距离为40，在索尼ce0682上发现小于40时会变成点击 三星9300上发现有时60像素以下滑动不起来
            m = (y1 + y2) // 2
            if y1 - y2 > 0:
                d = 30
            else:
                d = -30
            y1 = m + d  # TODO:坐标合法性判断
            y2 = m - d
        for _ in range(3):
            try:
                result = self.send_command(EnumCommand.CmdDrag, X1=x1, Y1=y1, X2=x2, Y2=y2,
                                           StepCount=count,
                                           SleepTime=wait_time,
                                           SendDownEvent=send_down_event,
                                           SendUpEvent=send_up_event)
            except AndroidSpyError as e:
                if str(e).find('java.lang.SecurityException') >= 0:
                    logger.info('java.lang.SecurityException,current activity:%s' %
                                self._device_driver.get_current_activity())  # 检测是否有弹窗
                    # 有时操作过快会出现该问题
                    time.sleep(0.1)
                else:
                    raise e
            else:
                return True
        logger.error('drag (%s, %s, %s, %s) failed' % (x1, y1, x2, y2))
        return False

    def enable_soft_input(self, control, enable=False):
        '''启用/禁止软键盘
        '''
        self.send_command(EnumCommand.CmdEnableSoftInput,
                          Control=control, Enable=enable)

    def _close_activity(self, activity):
        '''关闭Activity
        '''
        result = self.send_command(
            EnumCommand.CmdCloseActivity, Activity=activity)
        return result['Result']

    def close_activity(self, activity):
        '''关闭Activity
        '''
        try:
            return self._close_activity(activity)
        except ProcessExitError:
            logger.warn('close_activity error: process %s not exist' %
                        self._process_name)
            return -3  # 进程退出

    def eval_script(self, control, frame_xpaths, script):
        '''执行JavaScript
        '''
        result = self.send_command(
            EnumCommand.CmdEvalScript, Control=control, Script=script, Frame=frame_xpaths)
        return result['Result']

    def get_static_field_value(self, class_name, field_name):
        '''获取类`class_name`中静态字段`field_name`的值
        '''
        result = self.send_command(
            EnumCommand.CmdGetStaticFieldValue, ClassName=class_name, FieldName=field_name)
        return result['Result']

    def get_object_field_value(self, control, field_name):
        '''获取对象中字段`field_name`的值
        '''
        result = self.send_command(
            EnumCommand.CmdGetObjectFieldValue, Control=control, FieldName=field_name)
        return result['Result']

    def call_static_method(self, class_name, method_name, control=None, ret_type='', *args):
        '''调用静态方法
        '''
        kwargs = {}
        if control:
            kwargs['Control'] = control
        result = self.send_command(EnumCommand.CmdCallStaticMethod, ClassName=class_name,
                                   Method=method_name, RetType=ret_type, Args=args, **kwargs)
        return result['Result']

    def call_object_method(self, control, inner_obj, method_name, ret_type='', *args):
        '''调用对象方法
        '''
        result = self.send_command(EnumCommand.CmdCallObjectMethod, Control=control,
                                   InnerObject=inner_obj, Method=method_name, RetType=ret_type, Args=args)
        return result['Result']

    def call_external_method(self, jar_path, cls_name, add_to_classloader=False, **kwds):
        '''调用外部jar包中的方法
        '''
        result = self.send_command(EnumCommand.CmdCallExternalMethod, JarPath=jar_path,
                                   ClsName=cls_name, AddToClassLoader=add_to_classloader, Args=kwds)
        return result['Result']

    def set_activity_popup(self, activity, popup=False):
        '''禁用某些Activity自动弹出
        '''
        result = self.send_command(
            EnumCommand.CmdSetActivityPopup, Activity=activity, Popup=popup)
        return result['Result']

    def _format_control_tree(self, result, indent=0):
        '''格式化控件树
        '''
        if not result:
            return ''
        id = result.pop('Id')
        padding = '+--' * indent + '+'
        ret = '%s%s (' % (padding, id)
        rect = result.pop('Rect')
        bounding_rect = '(%s, %s, %s, %s)' % (
            rect['Left'], rect['Top'], rect['Width'], rect['Height'])
        children = result.pop('Children')

        for key in result.keys():
            ret += '%s: %s, ' % (key, result[key])
        ret += 'BoundingRect: %s)\n' % bounding_rect
        for child in children:
            ret += self._format_control_tree(child, indent + 1)
        return ret

    def _get_control_tree(self, activity, index):
        '''获取控件树
        '''
        return self.send_command('GetControlTree', Activity=activity, Index=index)['Result']

    def get_control_tree(self, activity, index=-1):
        '''获取控件树
        '''
        result = self._get_control_tree(activity, index)
        if len(activity) > 0:
            return self._format_control_tree(result)
        else:
            output = ''
            for activity in result.keys():
                output += '%s:\n' % activity
                output += self._format_control_tree(result[activity])
            return output

    def set_thread_priority(self, priority):
        '''设置测试线程优先级

        :param priority: 要设置的优先级
        :type priority:  EnumThreadPriority
        '''
        self.send_command(EnumCommand.CmdSetThreadPriority, Priority=priority)

    def set_webview_debugging_enabled(self, control, enable=True):
        '''设置WebView调试状态

        :param control: WebView控件hashcode
        :type  control: int/long
        :param enable: 是否开启WebView调试
        :type  enable: bool
        '''
        self.send_command(
            EnumCommand.CmdSetWebViewDebuggingEnabled, Control=control, Enable=True)


if __name__ == '__main__':
    pass



# -*- coding:UTF-8 -*-
#
# Tencent is pleased to support the open source community by making QTA available.
# Copyright (C) 2016THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the BSD 3-Clause License (the "License"); you may not use this
# file except in compliance with the License. You may obtain a copy of the License at
#
# https://opensource.org/licenses/BSD-3-Clause
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.
#

'''封装ADB功能
'''

from __future__ import unicode_literals
from __future__ import print_function

import os
import re
import six
import subprocess
import sys
import threading
import time

from pkg_resources import iter_entry_points
from qt4a.androiddriver.adbclient import ADBClient
from qt4a.androiddriver.util import Singleton, Deprecated, logger, ThreadEx, TimeoutError, InstallPackageFailedError, PermissionError, is_int, utf8_encode, encode_wrap, enforce_utf8_decode

try:
    import _strptime  # time.strptime() is not thread-safed, so import _strptime first, otherwise it raises an AttributeError: _strptime_time
except:
    pass
cur_path = os.path.dirname(os.path.abspath(__file__))


def get_adb_path():
    if sys.platform == 'win32':
        sep = ';'
        file_name = 'adb.exe'
    else:
        sep = ':'
        file_name = 'adb'

    for root in os.environ.get('PATH').split(sep):
        adb_path = os.path.join(root, file_name)
        if os.path.exists(adb_path):  # 优先使用环境变量中指定的 adb
            return adb_path

    return os.path.join(cur_path, 'tools', 'adb', sys.platform, file_name)


adb_path = get_adb_path()


def is_adb_server_opend():
    '''判断ADB Server是否开启
    '''
    import socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        sock.bind(('localhost', 5037))
        sock.close()
        return False
    except:
        return True


class EnumRootState(object):
    '''设备Root状态
    '''
    Unknown = 0  # 未知
    NonRoot = 1  # 非Root
    AdbdRoot = 2  # adbd以Root权限执行（执行adb root后）
    SuRoot = 3  # 使用su进入Root


class IADBBackend(object):
    '''ADBBackend接口定义
    '''

    @staticmethod
    def list_device():
        '''枚举设备列表
        '''
        pass

    @staticmethod
    def open_device(name):
        '''打开指定设备

        :param name: 设备名称
        :type  name: str
        :return: IADBBackend实例
        '''
        pass

    @property
    def device_host(self):
        '''设备主机
        '''
        pass

    @property
    def device_name(self):
        '''设备名
        '''
        pass

    def run_adb_cmd(self, cmd, *args, **kwargs):
        '''执行adb命令
        '''
        pass


class LocalADBBackend(IADBBackend):
    '''本地ADBBackend
    '''

    @staticmethod
    def start():
        if is_adb_server_opend():
            return False
        subprocess.call([adb_path, 'start-server'])
        return True

    @staticmethod
    def close():
        subprocess.call([adb_path, 'kill-server'])

    @staticmethod
    def list_device(device_host='127.0.0.1'):
        '''枚举设备列表
        '''
        if not is_adb_server_opend():
            return []
        result = ADBClient.get_client(device_host).call(
            'devices', retry_count=3)[0]
        result = result.split('\n')
        device_list = []
        for device in result:
            if len(device) <= 1 or not '\t' in device:
                continue
            device_name, status = device.split('\t')
            if status != 'device':
                continue
            device_list.append(device_name)
        return device_list

    @staticmethod
    def open_device(name):
        '''打开指定设备

        :param name: 设备名称
        :type  name: str
        :return: IADBBackend实例
        '''
        device_host = '127.0.0.1'
        if ':' in name:
            pattern = re.compile(r'^\d{3,5}$')
            pos = name.find(':')
            hostname = name[:pos]
            if not pattern.match(name[pos + 1:]):
                # adb connect device
                name = name[pos + 1:]
                device_host = hostname

        if name not in LocalADBBackend.list_device(device_host):
            raise RuntimeError('Device %s not exist in host %s' %
                               (name, device_host))

        return LocalADBBackend(device_host, name)

    def __init__(self, device_host, device_name, port=5037):
        self._device_host = device_host
        self._device_host_port = port
        self._device_name = device_name
        self._adb_client = ADBClient.get_client(
            self._device_host, self._device_host_port)

    @property
    def device_host(self):
        '''设备主机
        '''
        return self._device_host

    @property
    def device_name(self):
        '''设备名
        '''
        return self._device_name

    def run_adb_cmd(self, cmd, *args, **kwargs):
        '''执行adb命令
        '''
        timeout = kwargs.pop('timeout')
        sync = kwargs.pop('sync')
        return self._adb_client.call(cmd, self._device_name, *args, sync=sync, retry_count=1, timeout=timeout)


def static_result(func):
    '''固定返回结果函数
    '''

    def _wrap_func(self):
        attr = '_%s_result' % func.__name__
        if not hasattr(self, attr):
            result = func(self)
            setattr(self, attr, result)
        return getattr(self, attr)
    return _wrap_func


class ADB(object):
    '''封装ADB功能
    '''
    armeabi = 'armeabi'
    x86 = 'x86'

    connect_timeout = 300  # 连接设备的超时时间

    def __init__(self, backend):
        self._backend = backend
        self._device_name = self._backend.device_name
        self._root_state = EnumRootState.Unknown

        self._need_quote = None  # 执行shell命令时有些手机需要引号，有些不需要
        self._log_filter_thread_list = []  # 不打印log的线程id列表
        self._shell_prefix = None  # 有些设备上会有固定输出
        self._logcat_callbacks = []
        self._newline = None  # 不同手机使用的换行会不同

    @property
    def device_host(self):
        '''设备主机名
        '''
        return self._backend.device_host

    @property
    def device_name(self):
        '''设备名
        '''
        return self._backend.device_name

    def add_no_log_thread(self, thread):
        '''添加线程到不打印日志线程列表
        '''
        if not thread.ident in self._log_filter_thread_list:
            self._log_filter_thread_list.append(thread.ident)

    def remove_no_log_thread(self, thread):
        '''移除不打印日志线程列表中指定线程
        '''
        if thread.ident in self._log_filter_thread_list:
            self._log_filter_thread_list.remove(thread.ident)

    def run_adb_cmd(self, cmd, *args, **kwargs):
        '''执行adb命令
        '''
        retry_count = 3  # 默认最多重试3次
        if 'retry_count' in kwargs:
            retry_count = kwargs.pop('retry_count')
        timeout = 20
        if 'timeout' in kwargs:
            timeout = kwargs.pop('timeout')
        sync = True
        if 'sync' in kwargs:
            sync = kwargs.pop('sync')

        for _ in range(retry_count):
            if not threading.current_thread().ident in self._log_filter_thread_list:
                logger.info('adb %s:%s %s %s' % (
                    self._backend.device_host, self._backend.device_name, cmd, ' '.join(args)))
            time0 = time.clock()
            try:
                result = self._backend.run_adb_cmd(
                    cmd, *args, sync=sync, timeout=timeout, **kwargs)
            except Exception as e:
                logger.exception('Exec adb %s failed: %s' % (cmd, e))
                continue

            if not isinstance(result, tuple):
                return result
            if not threading.current_thread().ident in self._log_filter_thread_list:
                logger.info('执行ADB命令耗时：%s' % (time.clock() - time0))
            out, err = result

            if err:
                if b'error: device not found' in err:
                    self.run_adb_cmd('wait-for-device', retry_count=1,
                                     timeout=self.connect_timeout)  # 等待设备连接正常
                    return self.run_adb_cmd(cmd, *args, **kwargs)
                return err
            if isinstance(out, (bytes, str)):
                out = out.strip()
            return out

    def run_shell_cmd(self, cmd_line, root=False, **kwds):
        '''运行shell命令

        :param cmd_line: 要运行的命令行
        :param root: 是否使用root权限
        '''
        if isinstance(cmd_line, bytes):
            cmd_line = cmd_line.decode('utf8')

        if not self._newline:
            result = self.run_adb_cmd('shell', 'echo "1\n2"')
            if b'\r\n' in result:
                self._newline = b'\r\n'
            else:
                self._newline = b'\n'

        binary_output = False
        if 'binary_output' in kwds:
            binary_output = kwds.pop('binary_output')

        def _handle_result(result):
            if not isinstance(result, (bytes, str)):
                return result
            if self._newline != b'\n':
                result = result.replace(self._newline, b'\n')

            if binary_output:
                return result
            else:
                result = result.decode('utf8')

            if self._shell_prefix != None and self._shell_prefix > 0:
                result = '\n'.join(result.split('\n')[self._shell_prefix:])
            if result.startswith('WARNING: linker:'):
                # 虚拟机上可能会有这种错误：WARNING: linker: libdvm.so has text relocations. This is wasting memory and is a security risk. Please fix.
                lines = result.split('\n')
                idx = 1
                while idx < len(lines):
                    if not lines[idx].startswith('WARNING: linker:'):
                        break
                    idx += 1
                return '\n'.join(lines[idx:]).strip()
            else:
                return result

        if root:
            need_su = True
            if self._root_state == EnumRootState.Unknown:
                self._root_state = self.get_root_state()
            if self._root_state == EnumRootState.AdbdRoot:
                need_su = False
            elif self._root_state == EnumRootState.NonRoot:
                raise RuntimeError('device is not rooted')

            if not need_su:
                return self.run_shell_cmd(cmd_line, **kwds)
            if self._need_quote == None:
                self._check_need_quote()
            if self._need_quote:
                cmd_line = 'su -c \'%s\'' % cmd_line
            else:
                orig_cmd_line = cmd_line
                if not '&&' in orig_cmd_line:
                    cmd_line = 'su -c %s' % cmd_line
                else:
                    result = []
                    for cmd_line in orig_cmd_line.split('&&'):
                        cmd_line = 'su -c %s' % cmd_line
                        result.append(_handle_result(self.run_adb_cmd('shell', '%s' % cmd_line, **kwds)))
                    return '\n'.join(result)

        return _handle_result(self.run_adb_cmd('shell', cmd_line, **kwds))

    def reboot(self, _timeout=180):
        '''重启手机'''
        try:
            self.run_adb_cmd('reboot', retry_count=1, timeout=30)
        except TimeoutError:
            # 使用强杀init进程方式重启手机
            self.kill_process(1)
            time.sleep(10)  # 等待手机重启
        if _timeout > 0:
            self.wait_for_boot_complete(_timeout)

    def wait_for_boot_complete(self, _timeout=180):
        '''等待手机启动完成'''
        # 手机重启完后 adbd Insecure 启动时会导致adb断开重连，qt4a框架己经实现了adb root权限功能，测试手机请不要安装 adbd Insecure
        import time
        print('等待手机启动完成...')
        self.run_adb_cmd('wait-for-device', timeout=_timeout)
        boot_complete = False
        attempts = 0
        wait_period = 5
        while not boot_complete and (attempts * wait_period) < _timeout:
            output = self.run_shell_cmd(
                "getprop sys.boot_completed", retry_count=1)
            output = output.strip()
            if output == "1":
                boot_complete = True
            else:
                time.sleep(wait_period)
                attempts += 1
        if not boot_complete:
            raise RuntimeError(
                "dev.bootcomplete 标志在  %s 秒后仍未设置，手机重启失败" % _timeout)

    def start_logcat(self, process_list=[], params=''):
        '''运行logcat进程
        :param process_list: 要捕获日志的进程名或进程ID列表，为空则捕获所有进程
        :type process_list:  list
        '''
        if not hasattr(self, '_start_count'):
            self._start_count = 0
        self._start_count += 1
        if self._start_count > 1:
            return
        logger.debug('[ADB] start logcat')
        self.run_shell_cmd('logcat -c ' + params)  # 清除缓冲区
        if not hasattr(self, '_log_list'):
            self._log_list = []
        self._logcat_running = True
        self._log_pipe = self.run_shell_cmd(
            'logcat -v threadtime ' + params, sync=False)

        # self._logcat_thread_func(process_list)

        self._logcat_thread = ThreadEx(
            target=self._logcat_thread_func, args=[process_list, params])
        self._logcat_thread.setDaemon(True)
        self._logcat_thread.start()
        self._log_filter_thread_list.append(self._logcat_thread.ident)

    def stop_logcat(self):
        '''停止logcat
        '''
        if not hasattr(self, '_start_count') or self._start_count <= 0:
            logger.warn('[ADB] logcat not start')
            return
        self._start_count -= 1
        if self._start_count > 0:
            return
        logger.debug('[ADB] stop logcat')
        self._logcat_running = False
        if hasattr(self, '_log_pipe'):
            if self._log_pipe.poll() == None:  # 判断logcat进程是否存在
                try:
                    self._log_pipe.terminate()
                except WindowsError as e:
                    logger.warn('terminate logcat process failed: %s' % e)

        if hasattr(self, '_logcat_thread'):
            if self._logcat_thread.ident in self._log_filter_thread_list:
                self._log_filter_thread_list.remove(self._logcat_thread.ident)
            else:
                logger.warn('%s not in %s' % (
                    self._logcat_thread.ident, self._log_filter_thread_list))

    def get_log(self, clear=True):
        '''获取已经保存的log
        '''
        if not hasattr(self, '_log_list'):
            return []
        result = self._log_list
        if clear:
            self._log_list = []
        return result

    def save_log(self, save_path):
        '''保存log
        '''
        if not hasattr(self, '_log_list'):
            return
        log_list = self.get_log()
        for i, log in enumerate(log_list):
            if isinstance(log, bytes):
                # 先编码为unicode
                for code in ['utf8', 'gbk']:
                    try:
                        log = log.decode(code)
                        break
                    except UnicodeDecodeError as e:
                        # logger.warn('decode with %s error: %s' % (code, e))
                        pass
                else:
                    log = repr(log)
            log_list[i] = log.encode('utf8') if not isinstance(
                log, bytes) else log
        f = open(save_path, 'wb')
        f.write(b'\n'.join(log_list))
        f.close()

    def add_logcat_callback(self, callback):
        '''添加logcat回调
        '''
        if not callback in self._logcat_callbacks:
            self._logcat_callbacks.append(callback)

    def remove_logcat_callback(self, callback):
        '''移除logcat回调
        '''
        if callback in self._logcat_callbacks:
            self._logcat_callbacks.remove(callback)

    def insert_logcat(self, process_name, year, month_day, timestamp, level, tag, tid, content):
        self._log_list.append(b'[%s] [%d-%s %s] %s/%s(%d): %s' % (utf8_encode(process_name),
                                                                 int(year), utf8_encode(month_day),
                                                                 utf8_encode(timestamp),
                                                                 utf8_encode(level),
                                                                 utf8_encode(tag),
                                                                 int(tid),
                                                                 utf8_encode(content)))
        pid = 0
        pattern = re.compile(r'^(.+)\((\d+)\)$')
        ret = pattern.match(process_name)
        if ret:
            process_name = ret.group(1)
            pid = int(ret.group(2))
        for callback in self._logcat_callbacks:
            callback(pid, process_name, '%s-%s' % (year, month_day),
                     timestamp, level, tag, int(tid), content)

    def _logcat_thread_func(self, process_list, params=''):
        '''获取logcat线程
        '''
        import re
        # pattern = re.compile(r'([A-Z])/([\w|.]+)\s*\(\s*(\d+)\):.+') #标准格式
        # pattern = re.compile(r'([\d|-]+)\s+([\d|:|\.]+)\s+(\d+)\s+(\d+)\s+(\w)\s+(\S+)\s*:\s+(.+)')  # [^:]
        # 会过滤掉只有内容和内容为空的情况：--------- beginning of /dev/log/main not match pattern；04-16 10:09:25.170  2183  2183 D AndroidRuntime:
        pattern = re.compile(
            r'([\d|-]+)\s+([\d|:|\.]+)\s+(\d+)\s+(\d+)\s+(\w)\s+(.*?)\s*:\s*(.*)')
        # Date Time PID TID Level Tag Content
        pid_dict = {}
        filter_pid_list = []  # 没有找到匹配进程的列表
        zygote_pid = 0  # zygote进程ID

        while self._logcat_running:
            log = self._log_pipe.stdout.readline()
            log = enforce_utf8_decode(log).strip()

            if not log:
                if self._log_pipe.poll() != None:
                    logger.debug('logcat进程：%s 已退出' % self._log_pipe.pid)
                    # 进程已退出
                    # TODO: 解决logcat重复问题
                    if not self._logcat_running:
                        logger.info('logcat线程停止运行')
                        return
                    self._log_pipe = self.run_shell_cmd(
                        'logcat -v threadtime ' + params, sync=False)
                else:
                    continue

            if 'beginning of main' in log or 'beginning of system' in log:
                continue

            ret = pattern.match(log)
            if not ret:
                logger.info('log: %s not match pattern' % log)
                continue
            tag = ret.group(6).strip()
            if tag in ['inject', 'dexloader', 'ActivityInspect', 'MethodHook', 'androidhook']:
                logger.info(log)  # 测试桩日志加入到qt4a日志中
                continue

            if tag in ['Web Console']:
                if ret.group(7).startswith('[ClickListener]'):
                    logger.info(log)  # WebView的控件点击信息
                    continue

            pid = int(ret.group(3))
            if pid in filter_pid_list:
                continue

            init_process_list = ['<pre-initialized>', 'zygote']

            if not pid in pid_dict.keys():
                for item in self.list_process():
                    if zygote_pid == 0 and item['proc_name'] == 'zygote' and item['ppid'] == 1:
                        # zygote父进程ID为1
                        zygote_pid = item['pid']

                    for init_process in init_process_list:
                        if item['pid'] in pid_dict and pid_dict[item['pid']].startswith(init_process) and not item['proc_name'].startswith(init_process):

                            for i in range(len(self._log_list) - 1, -1, -1):
                                # 修复之前记录的“<pre-initialized>”进程
                                pre_process_name = b'[%s(%d)]' % (
                                    init_process, item['pid'])
                                if not pre_process_name in self._log_list[i]:
                                    continue
                                if process_list:
                                    del_flag = True
                                    for process in process_list:
                                        if pid == process or item['proc_name'].startswith(process):
                                            # 替换为真实进程名
                                            self._log_list[i] = self._log_list[i].replace(
                                                pre_process_name, ('[%s(%d)]' % (item['proc_name'], item['pid'])))
                                            del_flag = False
                                            break
                                    if del_flag:
                                        # 不在需要记录的进程列表中
                                        del self._log_list[i]
                                else:
                                    # 直接替换
                                    self._log_list[i] = self._log_list[i].replace(
                                        pre_process_name, ('[%s(%d)]' % (item['proc_name'], item['pid'])))
                    pid_dict[item['pid']] = item['proc_name']
#                     if item['proc_name'] in init_process_list and item['pid'] != zygote_pid:
#                         pid_dict[item['pid']] += '(%d)' % item['pid']
                if not pid in pid_dict.keys():
                    filter_pid_list.append(pid)
                    continue

            found = False
            if not process_list:
                found = True  # 不指定进程列表则捕获所有进程
            else:
                for process in process_list:
                    if pid == process or (pid in pid_dict and (pid_dict[pid].startswith(process) or pid_dict[pid].startswith('<pre-initialized>')
                                                               or (pid_dict[pid].startswith('zygote') and pid != zygote_pid))):  # 进程初始化中
                        found = True
                        break

            if found:
                import datetime
                if not hasattr(self, '_year'):
                    self._year = datetime.date.today().year
                try:
                    self.insert_logcat('%s(%d)' % (pid_dict.get(pid), pid), self._year, ret.group(
                        1), ret.group(2), ret.group(5), ret.group(6), ret.group(4), ret.group(7))
                except:
                    logger.exception('Insert logcat failed: %r' % log)

    @static_result
    def get_root_state(self):
        '''获取Root状态
        '''
        if self.is_adbd_root():
            return EnumRootState.AdbdRoot
        result = self.run_shell_cmd('su -c id')
        if 'su: not found' in result:
            return EnumRootState.NonRoot
        elif 'uid=0(root)' in result:
            return EnumRootState.SuRoot
        return EnumRootState.NonRoot

    @static_result
    def is_adbd_root(self):
        '''adbd是否以root权限运行
        '''
        result = self.run_shell_cmd('id')
        logger.debug('is_adbd_root: %s' % result)
        return 'uid=0(root)' in result

    def is_rooted(self):
        return self.get_root_state() in (EnumRootState.AdbdRoot, EnumRootState.SuRoot)

    def _check_need_quote(self, timeout=20):
        '''
        '''
        cmd = "su -c 'ls -l /data/data'"  # 默认方式为加引号，避免有些手机上对于存在空格的命令容易出错
        # 联想S899T上发现不加引号返回结果为空
        result = self.run_shell_cmd(cmd, timeout=timeout)
        if result.find('com.android.phone') >= 0:
            self._need_quote = True
        else:
            logger.debug(result)
            self._need_quote = False

    def _set_system_writable(self):
        '''修改system分区可写
        '''
        output = self.run_shell_cmd('mount -o rw,remount /system', True)
        result = self.run_shell_cmd('mount')
        for line in result.split('\n'):
            if '/system' in line and 'rw,' in line:
                return True
        raise RuntimeError('remount /system failed: %s' % output.strip())

    def forward(self, port1, port2, type='tcp'):
        '''端口转发
        :param port1: PC上的TCP端口
        :type port1:  int
        :param port2: 手机上的端口或LocalSocket地址
        :type port2:  int或String
        :param type:  手机上的端口类型
        :type type:   String，LocalSocket地址使用“localabstract”
        '''
        while 1:
            ret = self.run_adb_cmd('forward', 'tcp:%d' %
                                   (port1), '%s:%s' % (type, port2))
            if not 'cannot bind socket' in ret and not 'cannot bind to socket' in ret:
                return port1
            port1 += 1

    def remove_forward(self, port):
        '''移除指定的端口映射
        '''
        return 'cannot remove listener' in self.run_adb_cmd('forward', '--remove', 'tcp:%d' % (port))

    def create_tunnel(self, addr, type='tcp'):
        '''直接创建与手机中socket服务端的连接
        '''
        sock = self.run_adb_cmd('create_tunnel', '%s:%s' % (type, addr))
        if sock == '':
            return None
        return sock

    def _push_file(self, src_path, dst_path):
        '''以指定身份拷贝文件到手机中
        '''
        result = self.run_adb_cmd('push', src_path, dst_path, timeout=None)
        if 'No space left on device' in result or 'No such file or directory' in result:
            # 如果源文件不存在不会执行到这里
            raise RuntimeError('设备存储空间不足')
        return result

    def push_file(self, src_path, dst_path, uid=None):
        '''以指定身份拷贝文件到手机中
        '''
        if six.PY2 and isinstance(dst_path, unicode):
            dst_path = dst_path.encode('utf8')
        file_size = 0
        for _ in range(3):
            file_size = os.path.getsize(src_path)  # 防止取到的文件大小不正确
            result = self._push_file(src_path, dst_path)
            if file_size == 0:
                logger.warn('文件大小为0')
                return result
            if ('%d' % file_size) in result:
                try:
                    _, file_list = self.list_dir(dst_path)
                    if len(file_list) == 0:
                        logger.warn('push file failed: file not exist')
                    elif file_list[0]['size'] != file_size:
                        logger.warn('push file failed: file size error, expect %d, actual is %d' % (
                            file_size, file_list[0]['size']))
                        self.delete_file(dst_path)
                    else:
                        logger.debug(repr(file_list[0]))
                        if uid:
                            self.chown(dst_path, uid, uid)
                        return result
                except RuntimeError as e:
                    err_msg = e.args[0]
                    if six.PY2 and (not isinstance(err_msg, unicode)):
                        err_msg = err_msg.decode('utf8')
                    logger.warn(err_msg)
            else:
                logger.warn('push file failed: %s' % result)
        raise RuntimeError('Push file [%d]%s to device [%r] failed: %s' % (
            file_size, src_path, self._device_name, result))

    def pull_file(self, src_path, dst_path):
        '''从手机中拉取文件
        '''
        result = self.run_adb_cmd('pull', src_path, dst_path, timeout=600)
        if 'failed to copy' in result:
            raise RuntimeError(result)
        if not 'bytes in' in result:
            logger.warn(repr(result))
            logger.debug(self.run_shell_cmd('ls -l %s' % src_path, True))
        return result

    @staticmethod
    def _get_package_name(apk_path):
        '''获取安装包名
        '''
        import zipfile
        from ._axmlparser import AXMLPrinter
        package_name = ''
        zf = zipfile.ZipFile(apk_path, mode='r')
        for i in zf.namelist():
            if i == "AndroidManifest.xml":
                printer = AXMLPrinter(zf.read(i))
                package_name = printer.get_xml_obj().getElementsByTagName('manifest')[
                    0].getAttribute('package')
                break
        if not package_name:
            raise RuntimeError('获取安装包中的包名信息失败')
        return package_name

    def _install_apk(self, apk_path, package_name, reinstall=False):
        '''
        '''
        if self.get_sdk_version() <= 19:
            timeout = 3 * 60
        else:
            timeout = 6 * 60  # TODO: 9100安装5.0系统后安装应用超过3分钟
        cmdline = 'pm install %s %s' % ('-r' if reinstall else '', apk_path)
        ret = ''
        for i in range(3):
            # 处理一些必然会失败的情况，如方法数超标之类的问题
            try:
                if not self.is_rooted():
                    # 通知QT4A助手开始监控应用安装
                    self.run_shell_cmd('am broadcast -a startInstallMonitor')
                    ret = self.run_shell_cmd(
                        cmdline, retry_count=1, timeout=timeout)
                else:
                    proc = self.run_shell_cmd(
                        cmdline, True, sync=False)  # 使用root权限安装
                    time0 = time.time()
                    close_popup_count = 0
                    while time.time() - time0 < timeout:
                        if proc.poll() != None:
                            ret = proc.communicate()[0]
                            break
                        elif time.time() - time0 > 10 and close_popup_count < 2:
                            # 有些系统上弹窗会出现很久，关掉弹窗可以避免超时
                            self.run_shell_cmd('input keyevent 4')
                            close_popup_count += 1
                        time.sleep(1)
                    else:
                        raise TimeoutError('Install package timeout')

                    if not b'Success' in ret:
                        logger.warn('install with root failed: %s' % ret)
                        if not b'INSTALL_' in ret.strip().split(b'\n')[-1]:
                            # 权限弹窗导致的安装失败
                            ret = self.run_as(
                                'system', cmdline, retry_count=1, timeout=timeout)

                logger.debug(ret)
                if b'Success' in ret:
                    return True, ret
                elif i > 1 and b'INSTALL_FAILED_ALREADY_EXISTS' in ret:
                    # 出现至少一次超时，认为安装完成
                    return True, 'Success'
                elif b'INSTALL_FAILED_ALREADY_EXISTS' in ret:
                    # 尝试覆盖安装
                    return self._install_apk(apk_path, package_name, True)
                elif b'INSTALL_PARSE_FAILED_NO_CERTIFICATES' in ret or b'INSTALL_PARSE_FAILED_UNEXPECTED_EXCEPTION' in ret:
                    if i >= 2:
                        return False, ret
                    time.sleep(10)
                    continue
                elif b'INSTALL_FAILED_UPDATE_INCOMPATIBLE' in ret:
                    # 强制卸载应用
                    self.uninstall_app(package_name)
                    return self._install_apk(apk_path, package_name, False)
                elif b'INSTALL_PARSE_FAILED_INCONSISTENT_CERTIFICATES' in ret or b'INSTALL_FAILED_DEXOPT' in ret:
                    # 必须卸载安装
                    if not reinstall:
                        return False, ret
                    self.uninstall_app(package_name)
                    return self._install_apk(apk_path, package_name, False)
                elif b'INSTALL_FAILED_INSUFFICIENT_STORAGE' in ret:
                    # 有可能是存在/data/app-lib/packagename-1目录导致的
                    for i in (1, 2):
                        dir_path = '/data/app-lib/%s-%d' % (package_name, i)
                        if 'No such file or directory' in self.run_shell_cmd('ls -l %s' % dir_path, True):
                            continue
                        else:
                            self.delete_folder(dir_path)
                            break
                    else:
                        return False, ret
                elif b'INSTALL_FAILED_UID_CHANGED' in ret or b'INSTALL_FAILED_INTERNAL_ERROR' in ret:
                    # /data/data目录下存在文件夹没有删除
                    dir_path = '/data/data/%s' % package_name
                    for _ in range(3):
                        # 防止删除没有成功
                        self.delete_folder(dir_path)
                        if b'No such file or directory' in self.run_shell_cmd('ls -l %s' % dir_path, True):
                            break
                    continue
                elif b'INSTALL_FAILED_CANCELLED_BY_USER' in ret:
                    # 一般是ROM需要手动确认安装，改用system权限安装
                    ret = self.run_shell_cmd(
                        'su system %s' % cmdline, timeout=timeout)
                    if b'Success' in ret:
                        return True, ret
                else:
                    return False, ret
            except TimeoutError as e:
                logger.warn('install app timeout: %r' % e)
        else:
            logger.warn('install app failed')
            ret = self.run_shell_cmd(cmdline, timeout=timeout)  # 改用非root权限安装
            logger.debug(ret)
            if b'Success' in ret or b'INSTALL_FAILED_ALREADY_EXISTS' in ret:
                return True, 'Success'

        return False, ret

    def install_apk(self, apk_path, reinstall=False):
        '''安装应用
        '''
        if not os.path.exists(apk_path):
            raise RuntimeError('APK: %s 不存在' % apk_path)
        package_name = self._get_package_name(apk_path)

        tmp_path = '/data/local/tmp/%s.apk' % package_name
        self.push_file(apk_path, tmp_path)

        if not reinstall:
            self.uninstall_app(package_name)  # 先卸载，再安装
            result = self._install_apk(tmp_path, package_name, reinstall)
        else:
            result = self._install_apk(tmp_path, package_name, reinstall)
        # logger.debug(result)

        if result[0] == False:
            if not b'Failure' in result[1]:
                # 一般这种情况都是由于adb server意外退出导致，此时安装过程还会继续
                logger.warn('install app: %r' % result[1])
                timeout = 30
                time0 = time.time()
                while time.time() - time0 < timeout:
                    # 等待应用安装完成
                    if self.get_package_path(package_name):
                        break
                    time.sleep(1)
                else:
                    result = self._install_apk(
                        tmp_path, package_name, reinstall)
            else:
                err_msg = result[1]
                if six.PY2:
                    if isinstance(err_msg, unicode):
                        err_msg = err_msg.encode('utf8')
                    if isinstance(package_name, unicode):
                        package_name = package_name.encode('utf8')
                raise InstallPackageFailedError(
                    '安装应用%s失败：%s' % (package_name, err_msg))
        try:
            self.delete_file('/data/local/tmp/*.apk')
        except TimeoutError:
            pass

    def uninstall_app(self, pkg_name):
        '''卸载应用
        '''
        result = ''
        if not self.get_package_path(pkg_name):
            return True
        for _ in range(5):
            try:
                result = self.run_adb_cmd(
                    'uninstall', pkg_name, retry_count=1, timeout=30)
                break
            except RuntimeError:
                logger.exception('uninstall %s failed' % pkg_name)
                time.sleep(10)
        else:
            raise
        logger.debug('uninstall %s result: %r' % (pkg_name, result))
        if self.is_rooted():
            # 清理卸载可能遗留的cache文件
            cpu_abi = 'arm'
            if self.get_cpu_abi() == 'x86':
                cpu_abi = 'x86'  # TODO: 支持64位CPU
            self.delete_file('/data/dalvik-cache/%s/data@app@%s-*' %
                             (cpu_abi, pkg_name))
        return 'Success' in result

# ifndef __RELEASE__
    @Deprecated('uninstall_app')
    def uninstall_apk(self, pkg_name):
        '''卸载应用
        '''
        return self.uninstall_app(pkg_name)
# endif

    @encode_wrap
    def get_package_path(self, pkg_name):
        '''获取应用安装包路径
        '''
        for _ in range(3):
            # 为避免某些情况下获取不到应用安装包路径，重试多次
            result = self.run_shell_cmd('pm path %s' % pkg_name)
            logger.debug('get_package_path: %r' % result)
            pos = result.find('package:')
            if pos >= 0:
                return result[pos + 8:]
            time.sleep(1)
        return ''

    @encode_wrap
    def get_package_version(self, pkg_name):
        '''获取应用版本
        '''
        result = self.run_shell_cmd('dumpsys package %s' % pkg_name)
        for line in result.split('\n'):
            line = line.strip()
            if line.startswith('versionName='):
                return line[12:]

    @encode_wrap
    def _build_intent_extra_string(self, extra):
        '''构造intent参数列表
        '''
        extra_str = ''
        for key in extra:  # 指定额外参数
            p_type = ''
            value = extra[key]
            if isinstance(value, bytes):
                value = value.decode('utf8')

            if value in ['true', 'false']:
                p_type = 'z'  # EXTRA_BOOLEAN_VALUE
            elif isinstance(value, int):
                if is_int(value):
                    p_type = 'i'  # EXTRA_INT_VALUE
                else:
                    p_type = 'l'  # EXTRA_LONG_VALUE
            elif isinstance(value, float):
                p_type = 'f'  # EXTRA_FLOAT_VALUE
            elif value.startswith('file://'):  # EXTRA_URI_VALUE
                p_type = 'u'
            param = '-e%s %s %s ' % (p_type, key,
                                     ('"%s"' % value) if not p_type else value)
            if p_type:
                param = u'-' + param
            extra_str += param
        if len(extra_str) > 0:
            extra_str = extra_str[:-1]
        return extra_str

    @encode_wrap
    def start_activity(self, activity_name, action='', type='', data_uri='', extra={}, wait=True):
        '''打开一个Activity
        Warning: Activity not started, intent has been delivered to currently running top-most instance.
        Status: ok
        ThisTime: 0
        TotalTime: 0
        WaitTime: 2
        Complete
        '''
        if activity_name:
            activity_name = '-n %s' % activity_name
        if action:  # 指定Action
            action = '-a %s ' % action
        if type:
            type = '-t %s ' % type
        if data_uri:
            data_uri = '-d "%s" ' % data_uri
        extra_str = self._build_intent_extra_string(extra)
        W = u''
        if wait:
            W = '-W'  # 等待启动完成才返回
        # 如果/sbin/sh指向busybox，就会返回“/sbin/sh: am: not found”错误
        # 返回am找不到是因为am缺少“#!/system/bin/sh”
        command = 'am start %s %s %s%s%s%s' % (
            W, activity_name, action, type, data_uri, extra_str)
        if command[-1] == ' ':
            command = command[:-1]
        result = self.run_shell_cmd(command, timeout=15, retry_count=3)
        if 'Permission Denial' in result or (wait and (not 'Activity:' in result or not 'Complete' in result)):
            # 使用root权限运行
            if self.is_rooted():
                result = self.run_shell_cmd(
                    command, True, timeout=15, retry_count=3)
            else:
                package_name = activity_name.split('/')[0].split()[1]
                result = self.run_as(package_name, command,
                                     timeout=15, retry_count=3)
            # raise RuntimeError('打开Activity失败：\n%s' % result)
        if 'startActivityAndWait asks to run as user -2 but is calling from user 0' in result:
            command += ' --user 0'
            result = self.run_as(package_name, command,
                                 timeout=15, retry_count=3)
            logger.info('start activity command:%s' % command)
        if 'Permission Denial' in result or ('run as user -2 but is calling from user 0' in result) or (wait and not 'Complete' in result):
            raise RuntimeError('start activity failed: %s' % result)

        ret_dict = {}
        for line in result.split('\n'):
            if ': ' in line:
                key, value = line.split(': ')
                ret_dict[key] = value
        if 'Error' in ret_dict:
            raise RuntimeError(ret_dict['Error'])
        return ret_dict

    def start_service(self, service_name, extra={}):
        '''启动服务
        '''
        extra_str = self._build_intent_extra_string(extra)
        command = 'am startservice -n %s %s' % (service_name, extra_str)
        if command[-1] == ' ':
            command = command[:-1]
        result = self.run_shell_cmd(command, timeout=15, retry_count=3)
        if 'no service started' in result or 'java.lang.SecurityException' in result:
            raise RuntimeError('start service %s failed: %s' %
                               (service_name, result))

    def stop_service(self, service_name):
        '''停止服务
        '''
        result = self.run_shell_cmd(
            'am stopservice -n %s' % service_name, timeout=15, retry_count=3)
        if not 'Service stopped' in result and not 'was not running' in result:
            raise RuntimeError('stop service failed: %s' % result)

    def send_broadcast(self, action, extra={}):
        '''发送广播

        :param action: 广播使用的ACTION
        :type  action: string
        :param extra:  额外参数
        :type  extra:  dict
        '''
        extra_str = self._build_intent_extra_string(extra)
        command = 'am broadcast -a %s %s' % (action, extra_str)
        result = self.run_shell_cmd(command)
        if not 'Broadcast completed: result=0' in result:
            raise RuntimeError('Send broadcast failed: %s' % result)

    def get_property(self, prop):
        '''读取属性
        '''
        return self.run_shell_cmd('getprop %s' % prop)

    def set_property(self, prop, value):
        '''设置属性
        '''
        self.run_shell_cmd('setprop %s %s' % (prop, value), self.is_rooted())

    @static_result
    def get_cpu_abi(self):
        '''获取系统的CPU架构信息
        '''
        ret = self.run_shell_cmd('getprop ro.product.cpu.abi')
        if not ret:
            ret = 'armeabi'  # 有些手机可能没有这个系统属性
        return ret

    @static_result
    def get_device_model(self):
        '''获取设备型号
        '''
        model = self.run_shell_cmd('getprop ro.product.model')
        brand = self.run_shell_cmd('getprop ro.product.brand')
        if model.find(brand) >= 0:
            return model
        return '%s %s' % (brand, model)

    @static_result
    def get_system_version(self):
        '''获取系统版本
        '''
        return self.run_shell_cmd('getprop ro.build.version.release')

    @static_result
    def get_sdk_version(self):
        '''获取SDK版本
        '''
        return int(self.run_shell_cmd('getprop ro.build.version.sdk'))

    def get_uid(self, app_name):
        '''获取APP的uid
        '''
        result = self.run_shell_cmd('ls -l /data/data', True)
        for line in result.split('\n'):
            items = line.split(' ')
            for item in items:
                if not item:
                    continue
                if item == app_name:
                    return items[1]
        return None

    def is_selinux_opened(self):
        '''selinux是否是enforcing状态
        '''
        if self.get_sdk_version() < 18:
            return False
        return 'Enforcing' in self.run_shell_cmd('getenforce', True)

    def close_selinux(self):
        '''关闭selinux
        '''
        result = self.run_shell_cmd('setenforce 0', True)
        if 'Permission denied' in result:
            return False
        return True

    def chmod(self, file_path, attr):
        '''修改文件/目录属性

        :param file_path: 文件/目录路径
        :type file_path:  string
        :param attr:      设置的属性值，如：777
        :type attr:       int
        '''
        def _parse(num):
            num = str(num)
            attr = ''
            su_flag = ''
            if len(num) == 4:
                su_flag = int(num[0])
                num = num[1:]
            for c in num:
                c = int(c)
                if c & 4:
                    attr += 'r'
                else:
                    attr += '-'
                if c & 2:
                    attr += 'w'
                else:
                    attr += '-'
                if c & 1:
                    attr += 'x'
                else:
                    attr += '-'

            if su_flag and su_flag == 4:
                attr = attr[:2] + 's' + attr[3:]
            return attr

        ret = self.run_shell_cmd('chmod %s %s' %
                                 (attr, file_path), self.is_rooted())
        dir_list, file_list = self.list_dir(file_path)

        if len(dir_list) == 0 and len(file_list) == 1 and file_path.endswith('/' + file_list[0]['name']):
            # 这是一个文件
            new_attr = file_list[0]['attr']
        else:
            # 目录
            dir_name = file_path.split('/')[-1]
            parent_path = '/'.join(file_path.split('/')[:-1])
            dir_list, _ = self.list_dir(parent_path)
            for dir in dir_list:
                if dir['name'] == dir_name:
                    new_attr = dir['attr']
                    break

        if new_attr != _parse(attr):
            logger.warn('chmod failed: %r(%s)' % (ret, new_attr))
            return self.chmod(file_path, attr)
        return new_attr

    def chown(self, file_path, uid, gid):
        '''修改文件的拥有者和群组

        :param file_path: 文件路径
        :type file_path:  string
        :param uid:       拥有者
        :type uid:        string
        :param gid:       群组
        :type gid:        string
        '''
        self.run_shell_cmd('chown %s:%s %s' % (uid, gid, file_path), True)

    def mkdir(self, dir_path, mod=None):
        '''创建目录
        '''
        cmd = 'mkdir %s' % (dir_path)
        ret = self.run_shell_cmd(cmd, self.is_rooted())
#        if not 'File exists' in ret:
#            #加了-p参数貌似不会返回这个提示信息
        try:
            self.list_dir(dir_path)
        except RuntimeError as e:
            logger.warn('mkdir %s failed: %s(%s)' % (dir_path, ret, e))
            return self.mkdir(dir_path, mod)
        # 修改权限
        if mod != None:
            self.chmod(dir_path, mod)

    def list_dir(self, dir_path):
        '''列取目录
        '''
        if ' ' in dir_path:
            dir_path = '"%s"' % dir_path
        use_root = self.is_rooted()
        if use_root and dir_path.startswith('/sdcard') or dir_path.startswith('/storage/') or dir_path.startswith('/mnt/'):
            # 部分手机上发现用root权限访问/sdcard路径不一致
            use_root = False
        result = self.run_shell_cmd('ls -l %s' % dir_path, use_root)

        if 'Permission denied' in result:
            raise PermissionError(result)
        if 'No such file or directory' in result:
            raise RuntimeError('file or directory %s not exist' % dir_path)
        if 'Not a directory' in result:
            raise RuntimeError(u'%s %s' % (dir_path, result))

        dir_list = []
        file_list = []

        def _handle_name(name):
            return name.split('/')[-1]

        is_toybox = self.get_sdk_version() >= 24
        is_busybox = None
        # busybox格式 -rwxrwxrwx    1 shell    shell        13652 Jun  3 10:56 /data/local/tmp/qt4a/inject

        for line in result.split('\n'):
            items = line.split()
            if len(items) < 6:
                continue  # (6, 7, 9)
            if not line[0] in ('-', 'd', 'l'):
                continue

            is_dir = items[0][0] == 'd'  # 是否是目录
            is_link = items[0][0] == 'l'  # 软链
            if is_busybox == None:
                if is_toybox:
                    item = items[5]  # 日期字段
                else:
                    item = items[4]  # 日期字段
                    if is_dir or is_link:
                        item = items[3]  # 目录和软链没有size字段
                pattern = re.compile(r'\d{4}-\d{2}-\d{2}')
                if pattern.match(item):
                    is_busybox = False
                else:
                    is_busybox = True

            if not is_busybox:
                # 防止文件名称中有空格
                if not is_toybox:
                    if not is_dir and not is_link and len(items) > 7:
                        items[6] = line[line.find(items[6]):].strip()
                    elif is_dir and len(items) > 6:
                        items[5] = line[line.find(items[5]):].strip()
                else:
                    if not is_dir and not is_link and len(items) > 8:
                        items[7] = line[line.find(items[7]):].strip()
                    elif is_dir and len(items) > 7:
                        items[6] = line[line.find(items[6]):].strip()

            attrs = items[0]
            if attrs[0] == 'd':
                if is_busybox:
                    name = _handle_name(items[8])
                elif is_toybox:
                    name = items[7]
                else:
                    name = items[5]
                dir_list.append({'name': name, 'attr': attrs[1:]})
            elif attrs[0] == '-':
                if is_busybox:
                    name = _handle_name(items[8])
                    size = int(items[4])
                    last_modify_time = items[7]
                elif is_toybox:
                    name = _handle_name(items[7])
                    size = int(items[4])
                    last_modify_time = time.strptime(
                        '%s %s:00' % (items[5], items[6]), "%Y-%m-%d %X")
                else:
                    name = items[6]
                    size = int(items[3])
                    try:
                        last_modify_time = time.strptime(
                            '%s %s:00' % (items[4], items[5]), "%Y-%m-%d %X")
                    except:
                        # TODO: 即将删掉，调试用
                        logger.info('line=%s' % line)
                        raise
                file_list.append(
                    {'name': name, 'attr': attrs[1:], 'size': size, 'last_modify_time': last_modify_time})
            elif attrs[0] == 'l':  # link
                if is_busybox:
                    name = _handle_name(items[8])
                    last_modify_time = items[7]
                    link = items[10]
                elif is_toybox:
                    name = items[7]
                    last_modify_time = time.strptime(
                        '%s %s:00' % (items[5], items[6]), "%Y-%m-%d %X")
                    link = items[9]
                else:
                    name = items[5]
                    last_modify_time = time.strptime(
                        '%s %s:00' % (items[3], items[4]), "%Y-%m-%d %X")
                    link = items[7]
                file_list.append(
                    {'name': name, 'attr': attrs[1:], 'link': link, 'last_modify_time': last_modify_time})

        return dir_list, file_list

    def get_sdcard_path(self):
        '''获取sdcard路径
        '''
        path = '/sdcard'
        while True:
            dir_list, file_list = self.list_dir(path)
            if len(dir_list) == 0 and len(file_list) == 1 and 'link' in file_list[0]:
                # another link
                path = file_list[0]['link']
            else:
                break
        return path

    def get_file_info(self, file_path):
        '''获取文件信息
        '''
        return self.list_dir(file_path)[1][0]

    def copy_file(self, src_path, dst_path):
        '''在手机上拷贝文件
        '''
        if not hasattr(self, '_has_cp'):
            self._has_cp = 'not found' not in self.run_shell_cmd('cp')
        if self._has_cp:  # 不是所有的ROM都有cp命令
            self.run_shell_cmd('cp %s %s' %
                               (src_path, dst_path), self.is_rooted())
        else:
            self.run_shell_cmd('cat %s > %s' % (
                src_path, dst_path), self.is_rooted(), timeout=30)  # 部分手机上发现此方法耗时较多

    def delete_file(self, file_path):
        '''删除手机上文件
        '''
        if '*' in file_path:
            # 使用通配符时不能使用引号
            self.run_shell_cmd('rm -f %s' % file_path, self.is_rooted())
        else:
            file_path = file_path.replace('"', r'\"')
            self.run_shell_cmd('rm -f "%s"' % file_path, self.is_rooted())

    def delete_folder(self, folder_path):
        '''删除手机上的目录
        '''
        folder_path = folder_path.replace('"', r'\"')
        self.run_shell_cmd('rm -R "%s"' % folder_path, self.is_rooted())

    def run_as_by_app(self, package_name, cmdline, **kwargs):
        '''在app中执行命令
        '''
        cmd_res_path = '/sdcard/qt4a_cmd_res.txt'
        self.delete_file(cmd_res_path)
        timeout = 30
        if 'timeout' in kwargs:
            timeout = kwargs['timeout']
        try:
            self.start_activity('%s/com.test.androidspy.inject.CmdExecuteActivity' %
                                package_name, extra={'cmdline': cmdline, 'timeout': timeout}, wait=False)
        except Exception as e:
            if 'com.test.androidspy.inject.CmdExecuteActivity} does not exist' in e.args[0]:
                raise RuntimeError(
                    '该命令需要对apk重打包才能执行，请使用`qt4a-manage repack-apk -p /path/to/apk`命令进行重打包并安装后重试！')
            raise
        cmd_argv_list = cmdline.split()
        if len(cmd_argv_list) > 1 and cmd_argv_list[0] == 'pm' and cmd_argv_list[1] == 'clear':
            logger.info('run cmd:%s,return Success' % cmdline)
            time.sleep(2)
            return 'Success'

        time0 = time.time()
        while time.time() - time0 < timeout:
            try:
                self.list_dir(cmd_res_path)
                result = self.run_shell_cmd("cat %s" % cmd_res_path)
                return result
            except RuntimeError as e:
                logger.info('run_as_by_app exception:%s' % e)
                time.sleep(1)
        raise TimeoutError("run_as_by_app timeout:%d" % timeout)

    def run_as(self, package_name, cmdline, **kwargs):
        '''以package_name权限执行命令
        '''
        if self.is_rooted():
            if self._need_quote:
                cmdline = '"%s"' % cmdline
            cmdline = 'su %s %s' % (package_name, cmdline)
            return self.run_shell_cmd(cmdline, False, **kwargs)

        if ':' in package_name:
            package_name = package_name.split(':')[0]  # 允许传入进程名
        if '&&' in cmdline:
            cmndline = 'run-as %s sh -c "%s"' % (package_name, cmdline)
        else:
            cmndline = 'run-as %s %s' % (package_name, cmdline)
        result = self.run_shell_cmd(cmndline, **kwargs)
        run_as_succ = False
        if 'is unknown' in result:
            logger.info('Package %s not installed' % package_name)
        elif 'not debuggable' in result:
            logger.info('Package %s is not debuggable' % package_name)
        elif 'Could not set capabilities: Operation not permitted' in result:
            logger.info('Samsung device has bug with run-as command')
        elif 'run-as: exec failed for' in result:
            raise RuntimeError(result)
        else:
            run_as_succ = True
        if not run_as_succ:
            try:
                result = self.run_as_by_app(package_name, cmdline, **kwargs)
            except RuntimeError:
                logger.exception('run %s as %s by app failed' %
                                 (cmdline, package_name))
                raise PermissionError('run %s as %s failed' %
                                      (cmdline, package_name))
        return result

    def is_app_process64(self, process):
        '''是否是64位应用进程

        :param process: 进程名或进程ID
        :tytpe process: string/int
        '''
        process_name = ''
        if isinstance(process, six.string_types) and not process.isdigit():
            process_name = process
            pid = self.get_pid(process)
        else:
            pid = int(process)
        if pid <= 0:
            raise ValueError('process %s not exist' % process)
        if self.is_rooted():
            return 'app_process64' in self.run_shell_cmd('ls -l /proc/%d/exe' % pid, True)
        elif process_name:
            return 'app_process64' in self.run_as(process_name, 'ls -l /proc/%d/exe' % pid)
        else:
            raise ValueError('Non root device must pass process name')

    def _list_process(self):
        '''获取进程列表
        '''
        cmdline = 'ps'
        if self.get_sdk_version() >= 26:
            cmdline += ' -A'
        result = self.run_shell_cmd(cmdline).strip()
        lines = result.split('\n')
        busybox = False
        if lines[0].startswith('PID'):
            busybox = True

        result_list = []
        for i in range(1, len(lines)):
            lines[i] = lines[i].strip()
            if not lines[i]:
                continue
            items = lines[i].split()
            if not busybox:
                if len(items) < 9:
                    err_msg = "ps命令返回格式错误：\n%s" % lines[i]
                    if len(items) == 8:
                        result_list.append(
                            {'pid': int(items[1]), 'ppid': int(items[2]), 'proc_name': items[7]})
                    else:
                        raise RuntimeError(err_msg)
                else:
                    proc_name = items[8]
                    if len(proc_name) <= 1 and len(items) > 9:
                        proc_name = items[9]
                    result_list.append(
                        {'pid': int(items[1]), 'ppid': int(items[2]), 'proc_name': proc_name})
            else:
                idx = 4
                cmd = items[idx]
                if len(cmd) == 1:
                    # 有时候发现此处会有“N”
                    idx += 1
                    cmd = items[idx]
                idx += 1
                if cmd[0] == '{' and cmd[-1] == '}':
                    cmd = items[idx]
                ppid = 0
                if items[1].isdigit():
                    ppid = int(items[1])  # 有些版本中没有ppid
                result_list.append(
                    {'pid': int(items[0]), 'ppid': ppid, 'proc_name': cmd})
        return result_list

    def list_process(self):
        '''获取进程列表
        '''
        for _ in range(3):
            try:
                return self._list_process()
            except RuntimeError as e:
                logger.warn('%s' % e)
        else:
            raise RuntimeError('获取进程列表失败')

    def get_pid(self, proc_name):
        '''获取进程ID
        '''
        process_list = self.list_process()
        for process in process_list:
            if process['proc_name'] == proc_name:
                return process['pid']
        return 0

    def get_process_status(self, pid):
        '''获取进程状态信息
        '''
        ret = self.run_shell_cmd('cat /proc/%d/status' % pid, True)
        result = {}
        for line in ret.split('\n'):
            if not line:
                continue
            if not ':' in line:
                logger.warn('get_process_status line error: %r' % line)
                continue
            key, value = line.split(':')
            result[key] = value.strip()
        return result

    def get_process_user(self, pid):
        '''get procees user name

        :param pid: process id
        :type  pid: int
        '''
        uid = -1
        cmdline = 'cat /proc/%d/status' % pid
        result = self.run_shell_cmd(cmdline).strip()
        for line in result.split('\n'):
            line = line.strip()
            if line.startswith('Uid:'):
                uid = int(line.split()[1])
                break
        if uid < 0:
            raise RuntimeError('get uid of process %d failed' % pid)
        if uid < 10000:
            return uid
        cmdline = 'cat /proc/%d/cmdline' % pid
        result = self.run_shell_cmd(cmdline).strip().split('\x00')[0]
        if ':' in result:
            result = result.split(':')[0]
        return result

    def kill_process(self, proc_name_or_pid):
        '''杀进程
        '''
        kill_list = []
        package_name = None
        process_list = self.list_process()
        for process in process_list:
            if isinstance(proc_name_or_pid, six.string_types) and proc_name_or_pid in process['proc_name']:
                if process['proc_name'] == proc_name_or_pid:
                    # 保证主进程首先被杀
                    kill_list.insert(0, process['pid'])
                else:
                    kill_list.append(process['pid'])
            elif process['pid'] == proc_name_or_pid:
                kill_list.append(process['pid'])

        if not kill_list:
            return None  # 没有找到对应的进程
        if package_name == None and not self.is_rooted():
            package_name = self.get_process_user(kill_list[0])
        for i, pid in enumerate(kill_list):
            kill_list[i] = 'kill -9 %d' % pid
        cmd_line = ' && '.join(kill_list)

        if package_name == 2000:
            # shell process
            result = self.run_shell_cmd(cmd_line)
        elif self.is_rooted():
            result = self.run_shell_cmd(cmd_line, True)
        elif isinstance(package_name, six.string_types):
            # package
            result = self.run_as(package_name, cmd_line)
        else:
            raise PermissionError(
                'can\'t kill uid=%s process in non-root device' % package_name)

        if 'Operation not permitted' in result:
            raise PermissionError('run %s failed: %s' % (cmd_line, result))
        return True

    def get_device_imei(self):
        '''获取手机串号
        '''
        result = self.run_shell_cmd('dumpsys iphonesubinfo', self.is_rooted())
        for line in result.split('\n'):
            if line.find('Device ID') >= 0:
                return line.split('=')[1].strip()
        raise RuntimeError('获取imei号失败：%r' % result)

    def get_cpu_total_time(self):
        cpu_time = 0
        result = self.run_shell_cmd('cat /proc/stat')
        result = result.split('\n')[0]
        for item in result.split(' '):
            item = item.strip()
            if not item:
                continue
            if item == 'cpu':
                continue
            cpu_time += int(item)
        return cpu_time

    def get_process_cpu_time(self, pid):
        result = self.run_shell_cmd('cat /proc/%d/stat' % pid)
        result = result.split(' ')
        utime = int(result[13])
        stime = int(result[14])
        cutime = int(result[15])
        cstime = int(result[16])
        return utime + stime + cutime + cstime

    def get_thread_cpu_time(self, pid, tid):
        result = self.run_shell_cmd('cat /proc/%d/task/%d/stat' % (pid, tid))
        result = result.split(' ')
        utime = int(result[13])
        stime = int(result[14])
        cutime = int(result[15])
        cstime = int(result[16])
        return utime + stime + cutime + cstime

    def get_process_cpu(self, proc_name, interval=0.1):
        '''获取进程中每个线程的CPU占用率
        '''
        pid = self.get_pid(proc_name)
        # print (pid)
        if not pid:
            return None
        total_cpu1 = self.get_cpu_total_time()
        process_cpu1 = self.get_process_cpu_time(pid)
        thread_cpu1 = self.get_thread_cpu_time(pid, pid)
        time.sleep(interval)
        total_cpu2 = self.get_cpu_total_time()
        process_cpu2 = self.get_process_cpu_time(pid)
        thread_cpu2 = self.get_thread_cpu_time(pid, pid)
        total_cpu = total_cpu2 - total_cpu1
        process_cpu = process_cpu2 - process_cpu1
        thread_cpu = thread_cpu2 - thread_cpu1
        return process_cpu * 100 // total_cpu, thread_cpu * 100 // total_cpu

    @staticmethod
    def list_device():
        '''获取设备列表
        '''
        return LocalADBBackend.list_device()

    @staticmethod
    def is_local_device(device_id):
        '''是否是本地设备
        '''
        pattern = re.compile(r'([\w|\-|\.]+):(.+)')
        mat = pattern.match(device_id)
        if not mat or (mat.group(2).isdigit() and int(mat.group(2)) > 1024 and int(mat.group(2)) < 65536):
            return True
        else:
            return False

    @staticmethod
    def open_device(name_or_backend=None):
        '''打开设备
        '''
        if isinstance(name_or_backend, six.string_types):
            adb_backend = LocalADBBackend.open_device(name_or_backend)
        else:
            adb_backend = name_or_backend

        adb = ADB(adb_backend)

        if adb.is_rooted() and adb.is_selinux_opened():
            if not adb.close_selinux():
                logger.warn('Close selinux failed')
                # raise RuntimeError('关闭selinux失败，确认手机是否完美Root')
        return adb

    @staticmethod
    def connect_device(name):
        '''使用TCP连接设备
        '''
        proc = subprocess.Popen(
            [adb_path, 'connect', name], stdout=subprocess.PIPE)
        result = proc.stdout.read()
        if result.find('unable to connect to') >= 0:
            print(result, file=sys.stderr)
            return False
        return True

    def get_cpu_time(self):
        '''获取手机全局总时间片和空闲时间片
        '''
        import re
        cpu_time = 0
        result = self.run_shell_cmd('cat /proc/stat')
        result = result.split('\n')[0]
        result, num = re.subn(r'\s+', ' ', result)  # 将字符串中多个相连的空白字符合并成一个空白字符
        results = result.split(' ')
        if len(results) < 5:
            logger.warn('无法取得CPU时间片统计，请确保手机正常链接，并已启动！')
            return 0, 0
        idle_time = int(results[4])
        for item in results:
            item = item.strip()
            if not item:
                continue
            if item == 'cpu':
                continue
            cpu_time += int(item)
        return cpu_time, idle_time

    def get_cpu_usage(self, interval=0.5):
        '''获取手机全局CPU使用率
        '''
        total_time1, idle_time1 = self.get_cpu_time()
        time.sleep(interval)
        total_time2, idle_time2 = self.get_cpu_time()
        total_time = total_time2 - total_time1
        idle_time = idle_time2 - idle_time1
        if total_time == 0:
            return -1
        return (total_time - idle_time) * 100 // total_time

    @static_result
    def is_art(self):
        '''是否是art虚拟机
        '''
        ret = self.get_property('persist.sys.dalvik.vm.lib')
        if not ret:
            ret = self.get_property('persist.sys.dalvik.vm.lib.2')
        return 'libart.so' in ret

    def dump_stack(self, pid_or_procname):
        '''获取进程调用堆栈
        '''
        if isinstance(pid_or_procname, six.string_types):
            pid = self.get_pid(pid_or_procname)
        else:
            pid = pid_or_procname
        anr_dir = '/data/anr'
        try:
            self.list_dir(anr_dir)
        except RuntimeError:
            self.mkdir(anr_dir)
        self.chmod(anr_dir, 777)
        cmd = 'kill -3 %d' % pid
        self.run_shell_cmd(cmd, True)
        return self.run_shell_cmd('cat %s/traces.txt' % anr_dir, True)

    def get_state(self):
        '''获取设备状态
        '''
        return self.run_adb_cmd('get-state')


if __name__ == '__main__':
    pass




# -*- coding: utf-8 -*-

'''解析axml格式
copy from androguard
不要修改该文件
'''
from __future__ import print_function
import six
import sys
from struct import pack, unpack
from xml.sax.saxutils import escape
from xml.dom import minidom

TYPE_ATTRIBUTE = 2
TYPE_DIMENSION = 5
TYPE_FIRST_COLOR_INT = 28
TYPE_FIRST_INT = 16
TYPE_FLOAT = 4
TYPE_FRACTION = 6
TYPE_INT_BOOLEAN = 18
TYPE_INT_COLOR_ARGB4 = 30
TYPE_INT_COLOR_ARGB8 = 28
TYPE_INT_COLOR_RGB4 = 31
TYPE_INT_COLOR_RGB8 = 29
TYPE_INT_DEC = 16
TYPE_INT_HEX = 17
TYPE_LAST_COLOR_INT = 31
TYPE_LAST_INT = 31
TYPE_NULL = 0
TYPE_REFERENCE = 1
TYPE_STRING = 3

ATTRIBUTE_IX_NAMESPACE_URI = 0
ATTRIBUTE_IX_NAME = 1
ATTRIBUTE_IX_VALUE_STRING = 2
ATTRIBUTE_IX_VALUE_TYPE = 3
ATTRIBUTE_IX_VALUE_DATA = 4
ATTRIBUTE_LENGHT = 5

CHUNK_AXML_FILE = 0x00080003
CHUNK_RESOURCEIDS = 0x00080180
CHUNK_XML_FIRST = 0x00100100
CHUNK_XML_START_NAMESPACE = 0x00100100
CHUNK_XML_END_NAMESPACE = 0x00100101
CHUNK_XML_START_TAG = 0x00100102
CHUNK_XML_END_TAG = 0x00100103
CHUNK_XML_TEXT = 0x00100104
CHUNK_XML_LAST = 0x00100104

START_DOCUMENT = 0
END_DOCUMENT = 1
START_TAG = 2
END_TAG = 3
TEXT = 4

UTF8_FLAG = 0x00000100

RADIX_MULTS = [ 0.00390625, 3.051758E-005, 1.192093E-007, 4.656613E-010 ]
COMPLEX_UNIT_MASK = 15

def complexToFloat(xcomplex):
    return (float)(xcomplex & 0xFFFFFF00) * RADIX_MULTS[(xcomplex >> 4) & 3]

class StringBlock:
    def __init__(self, buff):
        self.start = buff.get_idx()
        self._cache = {}
        self.header = unpack('<h', buff.read(2))[0]
        self.header_size = unpack('<h', buff.read(2))[0]

        self.chunkSize = unpack('<i', buff.read(4))[0]
        self.stringCount = unpack('<i', buff.read(4))[0]
        self.styleOffsetCount = unpack('<i', buff.read(4))[0]

        self.flags = unpack('<i', buff.read(4))[0]
        self.m_isUTF8 = ((self.flags & UTF8_FLAG) != 0)

        self.stringsOffset = unpack('<i', buff.read(4))[0]
        self.stylesOffset = unpack('<i', buff.read(4))[0]

        self.m_stringOffsets = []
        self.m_styleOffsets = []
        self.m_strings = []
        self.m_styles = []

        for i in range(0, self.stringCount):
            self.m_stringOffsets.append(unpack('<i', buff.read(4))[0])

        for i in range(0, self.styleOffsetCount):
            self.m_styleOffsets.append(unpack('<i', buff.read(4))[0])

        size = self.chunkSize - self.stringsOffset
        if self.stylesOffset != 0:
            size = self.stylesOffset - self.stringsOffset

        # FIXME
        if (size % 4) != 0:
            print("ooo", file=sys.stderr)

        for i in range(0, size):
            self.m_strings.append(unpack('=b', buff.read(1))[0])

        if self.stylesOffset != 0:
            size = self.chunkSize - self.stylesOffset

            # FIXME
            if (size % 4) != 0:
                print("ooo", file=sys.stderr)

            for i in range(0, size // 4):
                self.m_styles.append(unpack('<i', buff.read(4))[0])

    def getString(self, idx):
        if idx in self._cache:
            return self._cache[idx]

        if idx < 0 or not self.m_stringOffsets or idx >= len(self.m_stringOffsets):
            return ""

        offset = self.m_stringOffsets[idx]

        if not self.m_isUTF8:
            length = self.getShort2(self.m_strings, offset)
            offset += 2
            self._cache[idx] = self.decode(self.m_strings, offset, length)
        else:
            offset += self.getVarint(self.m_strings, offset)[1]
            varint = self.getVarint(self.m_strings, offset)

            offset += varint[1]
            length = varint[0]

            self._cache[idx] = self.decode2(self.m_strings, offset, length)

        return self._cache[idx]

    def getStyle(self, idx):
        print (idx)
        print (idx in self.m_styleOffsets, self.m_styleOffsets[idx])

        print (self.m_styles[0])

    def decode(self, array, offset, length):
        length = length * 2
        length = length + length % 2

        data = b""

        for i in range(0, length):
            t_data = pack("=b", self.m_strings[offset + i])
            data += unicode(t_data, errors='ignore') if six.PY2 else t_data
            if data[-2:] == b"\x00\x00":
                break

        end_zero = data.find(b"\x00\x00")
        if end_zero != -1:
            data = data[:end_zero]

        return data.decode("utf-16", 'replace')

    def decode2(self, array, offset, length):
        data = b""

        for i in range(0, length):
            t_data = pack("=b", self.m_strings[offset + i])
            data += unicode(t_data, errors='ignore') if six.PY2 else t_data

        return data.decode("utf-8", 'replace')

    def getVarint(self, array, offset):
        val = array[offset]
        more = (val & 0x80) != 0
        val &= 0x7f

        if not more:
            return val, 1
        return val << 8 | array[offset + 1] & 0xff, 2

    def getShort(self, array, offset):
        value = array[offset // 4]
        if ((offset % 4) // 2) == 0:
            return value & 0xFFFF
        else:
            return value >> 16

    def getShort2(self, array, offset):
        return (array[offset + 1] & 0xff) << 8 | array[offset] & 0xff

    def show(self):
        print ("StringBlock", hex(self.start), hex(self.header), hex(self.header_size), hex(self.chunkSize), hex(self.stringsOffset), self.m_stringOffsets)
        for i in range(0, len(self.m_stringOffsets)):
            print (i, repr(self.getString(i)))

class SV :
    def __init__(self, size, buff) :
        self.__size = size
        self.__value = unpack(self.__size, buff)[0]

    def _get(self) :
        return pack(self.__size, self.__value)

    def __str__(self) :
        return "0x%x" % self.__value

    def __int__(self) :
        return self.__value

    def get_value_buff(self) :
        return self._get()

    def get_value(self) :
        return self.__value

    def set_value(self, attr) :
        self.__value = attr

class BuffHandle:
    def __init__(self, buff):
        self.__buff = buff
        self.__idx = 0

    def size(self):
        return len(self.__buff)

    def set_idx(self, idx):
        self.__idx = idx

    def get_idx(self):
        return self.__idx

    def readNullString(self, size):
        data = self.read(size)
        return data

    def read_b(self, size) :
        return self.__buff[ self.__idx : self.__idx + size ]

    def read_at(self, offset, size):
        return self.__buff[ offset : offset + size ]

    def read(self, size) :
        if isinstance(size, SV) :
            size = size.value

        buff = self.__buff[ self.__idx : self.__idx + size ]
        self.__idx += size

        return buff

    def end(self) :
        return self.__idx == len(self.__buff)

class AXMLParser:
    def __init__(self, raw_buff):
        self.reset()

        self.buff = BuffHandle(raw_buff)

        self.buff.read(4)
        self.buff.read(4)

        self.sb = StringBlock(self.buff)

        self.m_resourceIDs = []
        self.m_prefixuri = {}
        self.m_uriprefix = {}
        self.m_prefixuriL = []

        self.visited_ns = []

    def reset(self):
        self.m_event = -1
        self.m_lineNumber = -1
        self.m_name = -1
        self.m_namespaceUri = -1
        self.m_attributes = []
        self.m_idAttribute = -1
        self.m_classAttribute = -1
        self.m_styleAttribute = -1

    def next(self):
        self.doNext()
        return self.m_event

    def doNext(self):
        if self.m_event == END_DOCUMENT:
            return

        event = self.m_event

        self.reset()
        while True:
            chunkType = -1

            # Fake END_DOCUMENT event.
            if event == END_TAG:
                pass

            # START_DOCUMENT
            if event == START_DOCUMENT:
                chunkType = CHUNK_XML_START_TAG
            else:
                if self.buff.end():
                    self.m_event = END_DOCUMENT
                    break
                chunkType = unpack('<L', self.buff.read(4))[0]

            if chunkType == CHUNK_RESOURCEIDS:
                chunkSize = unpack('<L', self.buff.read(4))[0]
                # FIXME
                if chunkSize < 8 or chunkSize % 4 != 0:
                    print("ooo", file=sys.stderr)

                for i in range(0, chunkSize // 4 - 2):
                    self.m_resourceIDs.append(unpack('<L', self.buff.read(4))[0])

                continue

            # FIXME
            if chunkType < CHUNK_XML_FIRST or chunkType > CHUNK_XML_LAST:
                print("ooo", file=sys.stderr)

            # Fake START_DOCUMENT event.
            if chunkType == CHUNK_XML_START_TAG and event == -1:
                self.m_event = START_DOCUMENT
                break

            self.buff.read(4)  # /*chunkSize*/
            lineNumber = unpack('<L', self.buff.read(4))[0]
            self.buff.read(4)  # 0xFFFFFFFF

            if chunkType == CHUNK_XML_START_NAMESPACE or chunkType == CHUNK_XML_END_NAMESPACE:
                if chunkType == CHUNK_XML_START_NAMESPACE:
                    prefix = unpack('<L', self.buff.read(4))[0]
                    uri = unpack('<L', self.buff.read(4))[0]

                    self.m_prefixuri[prefix] = uri
                    self.m_uriprefix[uri] = prefix
                    self.m_prefixuriL.append((prefix, uri))
                    self.ns = uri
                else:
                    self.ns = -1
                    self.buff.read(4)
                    self.buff.read(4)
                    (prefix, uri) = self.m_prefixuriL.pop()
                    #del self.m_prefixuri[ prefix ]
                    #del self.m_uriprefix[ uri ]

                continue

            self.m_lineNumber = lineNumber

            if chunkType == CHUNK_XML_START_TAG:
                self.m_namespaceUri = unpack('<L', self.buff.read(4))[0]
                self.m_name = unpack('<L', self.buff.read(4))[0]

                # FIXME
                self.buff.read(4)  # flags

                attributeCount = unpack('<L', self.buff.read(4))[0]
                self.m_idAttribute = (attributeCount >> 16) - 1
                attributeCount = attributeCount & 0xFFFF
                self.m_classAttribute = unpack('<L', self.buff.read(4))[0]
                self.m_styleAttribute = (self.m_classAttribute >> 16) - 1

                self.m_classAttribute = (self.m_classAttribute & 0xFFFF) - 1

                for i in range(0, attributeCount * ATTRIBUTE_LENGHT):
                    self.m_attributes.append(unpack('<L', self.buff.read(4))[0])

                for i in range(ATTRIBUTE_IX_VALUE_TYPE, len(self.m_attributes), ATTRIBUTE_LENGHT):
                    self.m_attributes[i] = self.m_attributes[i] >> 24

                self.m_event = START_TAG
                break

            if chunkType == CHUNK_XML_END_TAG:
                self.m_namespaceUri = unpack('<L', self.buff.read(4))[0]
                self.m_name = unpack('<L', self.buff.read(4))[0]
                self.m_event = END_TAG
                break

            if chunkType == CHUNK_XML_TEXT:
                self.m_name = unpack('<L', self.buff.read(4))[0]

                # FIXME
                self.buff.read(4)
                self.buff.read(4)

                self.m_event = TEXT
                break

    def getPrefixByUri(self, uri):
        try:
            return self.m_uriprefix[uri]
        except KeyError:
            return -1

    def getPrefix(self):
        try:
            return self.sb.getString(self.m_uriprefix[self.m_namespaceUri])
        except KeyError:
            return u''

    def getName(self):
        if self.m_name == -1 or (self.m_event != START_TAG and self.m_event != END_TAG) :
            return u''

        return self.sb.getString(self.m_name)

    def getText(self) :
        if self.m_name == -1 or self.m_event != TEXT :
            return u''

        return self.sb.getString(self.m_name)

    def getNamespacePrefix(self, pos):
        prefix = self.m_prefixuriL[pos][0]
        return self.sb.getString(prefix)

    def getNamespaceUri(self, pos):
        uri = self.m_prefixuriL[pos][1]
        return self.sb.getString(uri)

    def getXMLNS(self):
        buff = ""
        for i in self.m_uriprefix:
            if i not in self.visited_ns:
                buff += "xmlns:%s=\"%s\"\n" % (self.sb.getString(self.m_uriprefix[i]), self.sb.getString(self.m_prefixuri[self.m_uriprefix[i]]))
                self.visited_ns.append(i)
        return buff

    def getNamespaceCount(self, pos) :
        pass

    def getAttributeOffset(self, index):
        # FIXME
        if self.m_event != START_TAG:
            print("Current event is not START_TAG.", file=sys.stderr)

        offset = index * 5
        # FIXME
        if offset >= len(self.m_attributes):
            print("Invalid attribute index", file=sys.stderr)

        return offset

    def getAttributeCount(self):
        if self.m_event != START_TAG:
            return -1

        return len(self.m_attributes) // ATTRIBUTE_LENGHT

    def getAttributePrefix(self, index):
        offset = self.getAttributeOffset(index)
        uri = self.m_attributes[offset + ATTRIBUTE_IX_NAMESPACE_URI]

        prefix = self.getPrefixByUri(uri)

        if prefix == -1:
            return ""

        return self.sb.getString(prefix)

    def getAttributeName(self, index) :
        offset = self.getAttributeOffset(index)
        name = self.m_attributes[offset + ATTRIBUTE_IX_NAME]

        if name == -1 :
            return ""

        return self.sb.getString(name)

    def getAttributeValueType(self, index) :
        offset = self.getAttributeOffset(index)
        return self.m_attributes[offset + ATTRIBUTE_IX_VALUE_TYPE]

    def getAttributeValueData(self, index) :
        offset = self.getAttributeOffset(index)
        return self.m_attributes[offset + ATTRIBUTE_IX_VALUE_DATA]

    def getAttributeValue(self, index) :
        offset = self.getAttributeOffset(index)
        valueType = self.m_attributes[offset + ATTRIBUTE_IX_VALUE_TYPE]
        if valueType == TYPE_STRING :
            valueString = self.m_attributes[offset + ATTRIBUTE_IX_VALUE_STRING]
            return self.sb.getString(valueString)
        # WIP
        return ""
        #int valueData=m_attributes[offset+ATTRIBUTE_IX_VALUE_DATA];
        #return TypedValue.coerceToString(valueType,valueData);

DIMENSION_UNITS = [ "px", "dip", "sp", "pt", "in", "mm" ]
FRACTION_UNITS = [ "%", "%p" ]

def long2int(l) :
    if l > 0x7fffffff :
        l = (0x7fffffff & l) - 0x80000000
    return l

class AXMLPrinter:
    def __init__(self, raw_buff):
        self.axml = AXMLParser(raw_buff)
        self.xmlns = False

        self.buff = u''

        while True:
            _type = self.axml.next()
#           print ("tagtype = ", _type)

            if _type == START_DOCUMENT:
                self.buff += u'<?xml version="1.0" encoding="utf-8"?>\n'
            elif _type == START_TAG:
                self.buff += u'<' + self.getPrefix(self.axml.getPrefix()) + self.axml.getName() + u'\n'
                self.buff += self.axml.getXMLNS()

                for i in range(0, self.axml.getAttributeCount()):
                    self.buff += "%s%s=\"%s\"\n" % (self.getPrefix(
                        self.axml.getAttributePrefix(i)), self.axml.getAttributeName(i), self._escape(self.getAttributeValue(i)))

                self.buff += u'>\n'

            elif _type == END_TAG :
                self.buff += "</%s%s>\n" % (self.getPrefix(self.axml.getPrefix()), self.axml.getName())

            elif _type == TEXT :
                self.buff += "%s\n" % self.axml.getText()

            elif _type == END_DOCUMENT :
                break

    # pleed patch
    def _escape(self, s):
        s = s.replace("&", "&amp;")
        s = s.replace('"', "&quot;")
        s = s.replace("'", "&apos;")
        s = s.replace("<", "&lt;")
        s = s.replace(">", "&gt;")
        return escape(s)

    def get_buff(self):
        return self.buff.encode('utf-8')

    def get_xml(self):
        return minidom.parseString(self.get_buff()).toprettyxml()

    def get_xml_obj(self):
        return minidom.parseString(self.get_buff())

    def getPrefix(self, prefix):
        if prefix == None or len(prefix) == 0:
            return u''

        return prefix + u':'

    def getAttributeValue(self, index):
        _type = self.axml.getAttributeValueType(index)
        _data = self.axml.getAttributeValueData(index)

        if _type == TYPE_STRING:
            return self.axml.getAttributeValue(index)

        elif _type == TYPE_ATTRIBUTE:
            return "?%s%08X" % (self.getPackage(_data), _data)

        elif _type == TYPE_REFERENCE:
            return "@%s%08X" % (self.getPackage(_data), _data)

        elif _type == TYPE_FLOAT:
            return "%f" % unpack("=f", pack("=L", _data))[0]

        elif _type == TYPE_INT_HEX:
            return "0x%08X" % _data

        elif _type == TYPE_INT_BOOLEAN:
            if _data == 0:
                return "false"
            return "true"

        elif _type == TYPE_DIMENSION:
            return "%f%s" % (complexToFloat(_data), DIMENSION_UNITS[_data & COMPLEX_UNIT_MASK])

        elif _type == TYPE_FRACTION:
            return "%f%s" % (complexToFloat(_data) * 100, FRACTION_UNITS[_data & COMPLEX_UNIT_MASK])

        elif _type >= TYPE_FIRST_COLOR_INT and _type <= TYPE_LAST_COLOR_INT:
            return "#%08X" % _data

        elif _type >= TYPE_FIRST_INT and _type <= TYPE_LAST_INT:
            return "%d" % long2int(_data)

        return "<0x%X, type 0x%02X>" % (_data, _type)

    def getPackage(self, id):
        if id >> 24 == 1:
            return "android:"
        return ""


if __name__ == '__main__':
    pass
